'''

.. testsetup::

    from pyroute2 import NDB
    from pyroute2 import config
    config.mock_iproute = True
    ndb = NDB()

.. testsetup:: tables

    from pyroute2 import NDB
    from pyroute2 import config
    config.mock_iproute = True
    ndb = NDB()
    ndb.routes.create(
        dst='1.1.1.1/32', gateway='127.0.0.10', oif=1, table=101
    ).commit()
    ndb.routes.create(
        dst='1.1.1.2/32', gateway='127.0.0.10', oif=1, table=5001
    ).commit()
    ndb.routes.create(
        dst='1.1.1.3/32', gateway='127.0.0.10', oif=1, table=5002
    ).commit()

.. testsetup:: metrics

    from pyroute2 import NDB
    from pyroute2 import config
    config.mock_iproute = True
    ndb = NDB()
    ndb.routes.create(
        dst='10.0.0.0/24', gateway='127.0.0.10'
    ).commit()

Simple routes
=============

Ordinary routes management is really simple:

.. testcode::

    # create a route
    ndb.routes.create(
        dst='10.0.0.0/24',
        gateway='192.168.122.1'
    ).commit()

    # retrieve a route and change it
    with ndb.routes['10.0.0.0/24'] as route:
        route.set(gateway='192.168.122.10')

    # remove a route
    with ndb.routes['10.0.0.0/24'] as route:
        route.remove()


Multiple routing tables
=======================

But Linux systems have more than one routing table:

.. doctest:: tables

    >>> set((x.table for x in ndb.routes.summary()))
    {101, 5001, 5002, 254, 255}

The main routing table is 254. All the routes people mostly work with are
in that table. To address routes in other routing tables, you can use dict
specs:

.. testcode::

    ndb.routes.create(
        dst='10.0.0.0/24',
        gateway='192.168.122.1',
        table=101
    ).commit()

    with ndb.routes[{'table': 101, 'dst': '10.0.0.0/24'}] as route:
        route.set('gateway', '192.168.122.10')
        route.set('priority', 500)

    with ndb.routes[{'table': 101, 'dst': '10.0.0.0/24'}] as route:
        route.remove()

Route metrics
=============

`route['metrics']` attribute provides a dictionary-like object that
reflects route metrics like hop limit, mtu etc:

.. testcode:: metrics

    # set up all metrics from a dictionary
    with ndb.routes['10.0.0.0/24'] as route:
        route.set('metrics', {'mtu': 1500, 'hoplimit': 20})

    # fix individual metrics
    with ndb.routes['10.0.0.0/24']['metrics'] as metrics:
        metrics.set('mtu', 1500)
        metrics.set('hoplimit', 20)

MPLS routes
===========

See here: :ref:`mpls`

'''

import json
import struct
import time
import uuid
from collections import OrderedDict
from functools import partial
from socket import AF_INET, inet_pton

from pyroute2.common import AF_MPLS, basestring
from pyroute2.netlink.rtnl.rtmsg import LWTUNNEL_ENCAP_MPLS, nh, rtmsg
from pyroute2.requests.common import MPLSTarget
from pyroute2.requests.route import RouteFieldFilter

from ..auth_manager import check_auth
from ..objects import RTNL_Object
from ..report import Record

_dump_rt = ['main.f_%s' % x[0] for x in rtmsg.sql_schema()][:-2]
_dump_nh = ['nh.f_%s' % x[0] for x in nh.sql_schema()][:-2]

F_RTA_MULTIPATH = 1
F_RTA_ENCAP = 2
F_RTA_METRICS = 4


def get_route_id(schema, target, event):
    keys = ['f_target = %s' % schema.plch]
    values = [target]
    for key in schema.indices['routes']:
        keys.append('f_%s = %s' % (key, schema.plch))
        values.append(event.get(key) or event.get_attr(key))
    #
    spec = 'WHERE %s' % ' AND '.join(keys)
    s_req = 'SELECT f_route_id FROM routes %s' % spec
    #
    # get existing route_id
    for route_id in schema.execute(s_req, values).fetchall():
        #
        # if exists
        return route_id[0][0]
    #
    # or create a new route_id
    return str(uuid.uuid4())


def load_rtmsg(schema, target, event):
    route_id = None
    post = []

    # fix RTA_TABLE
    rta_table = event.get_attr('RTA_TABLE', -1)
    if rta_table == -1:
        event['attrs'].append(['RTA_TABLE', 254])

    #
    # manage gc marks on related routes
    #
    # only for automatic routes:
    #   - table 254 (main)
    #   - proto 2 (kernel)
    #   - scope 253 (link)
    elif (
        (event.get_attr('RTA_TABLE') == 254)
        and (event['proto'] == 2)
        and (event['scope'] == 253)
        and (event['family'] == AF_INET)
    ):
        evt = event['header']['type']
        #
        # set f_gc_mark = timestamp for "del" events
        # and clean it for "new" events
        #
        try:
            rtmsg_gc_mark(
                schema, target, event, int(time.time()) if (evt % 2) else None
            )
        except Exception as e:
            schema.log.error('gc_mark event: %s' % (event,))
            schema.log.error('gc_mark: %s' % (e,))

    #
    # only for RTM_NEWROUTE events
    #
    if not event['header']['type'] % 2:
        event['deps'] = 0
        #
        # RTA_MULTIPATH
        #
        mp = event.get_attr('RTA_MULTIPATH')
        if mp:
            #
            # create key
            route_id = route_id or get_route_id(schema, target, event)
            #
            # load multipath
            for idx in range(len(mp)):
                mp[idx]['header'] = {}  # for load_netlink()
                mp[idx]['route_id'] = route_id  # set route_id on NH
                mp[idx]['nh_id'] = idx  # add NH number
                post.append(
                    partial(
                        schema.load_netlink, 'nh', target, mp[idx], 'routes'
                    )
                )
                event['deps'] |= F_RTA_MULTIPATH
        #
        # RTA_ENCAP
        #
        encap = event.get_attr('RTA_ENCAP')
        encap_type = event.get_attr('RTA_ENCAP_TYPE')
        if encap_type == LWTUNNEL_ENCAP_MPLS:
            route_id = route_id or get_route_id(schema, target, event)
            #
            encap['header'] = {}
            encap['route_id'] = route_id
            post.append(
                partial(
                    schema.load_netlink, 'enc_mpls', target, encap, 'routes'
                )
            )
            event['deps'] |= F_RTA_ENCAP
        #
        # RTA_METRICS
        #
        metrics = event.get_attr('RTA_METRICS')
        if metrics:
            #
            # create key
            route_id = route_id or get_route_id(schema, target, event)
            #
            metrics['header'] = {}
            metrics['route_id'] = route_id
            post.append(
                partial(
                    schema.load_netlink, 'metrics', target, metrics, 'routes'
                )
            )
            event['deps'] |= F_RTA_METRICS
    #
    if route_id is not None:
        event['route_id'] = route_id
    schema.load_netlink('routes', target, event)
    #
    for procedure in post:
        procedure()


def rtmsg_gc_mark(schema, target, event, gc_mark=None):
    #
    if gc_mark is None:
        gc_clause = ' AND f_gc_mark IS NOT NULL'
    else:
        gc_clause = ''
    #
    # select all routes for that OIF where f_gc_mark is not null
    #
    key_fields = ','.join(['f_%s' % x for x in schema.indices['routes']])
    key_query = ' AND '.join(
        ['f_%s = %s' % (x, schema.plch) for x in schema.indices['routes']]
    )
    routes = schema.execute(
        '''
                       SELECT %s,f_RTA_GATEWAY FROM routes WHERE
                       f_target = %s AND f_RTA_OIF = %s AND
                       f_RTA_GATEWAY IS NOT NULL %s AND
                       f_family = 2
                       '''
        % (key_fields, schema.plch, schema.plch, gc_clause),
        (target, event.get_attr('RTA_OIF')),
    ).fetchmany()
    #
    # get the route's RTA_DST and calculate the network
    #
    addr = event.get_attr('RTA_DST')
    net = struct.unpack('>I', inet_pton(AF_INET, addr))[0] & (
        0xFFFFFFFF << (32 - event['dst_len'])
    )
    #
    # now iterate all the routes from the query above and
    # mark those with matching RTA_GATEWAY
    #
    for route in routes:
        # get route GW
        gw = route[-1]
        try:
            gwnet = struct.unpack('>I', inet_pton(AF_INET, gw))[0] & net
            if gwnet == net:
                (
                    schema.execute(
                        'UPDATE routes SET f_gc_mark = %s '
                        'WHERE f_target = %s AND %s'
                        % (schema.plch, schema.plch, key_query),
                        (gc_mark, target) + route[:-1],
                    )
                )
        except Exception as e:
            schema.log.error('gc_mark event: %s' % (event,))
            schema.log.error('gc_mark: %s : %s' % (e, route))


rt_schema = (
    rtmsg.sql_schema()
    .push('route_id', 'TEXT UNIQUE')
    .push('gc_mark', 'INTEGER')
    .push('deps', 'INTEGER')
    .unique_index(
        'family',
        'dst_len',
        'tos',
        'scope',
        'type',
        'RTA_DST',
        'RTA_OIF',
        'RTA_PRIORITY',
        'RTA_TABLE',
        'RTA_VIA',
        'RTA_NEWDST',
    )
    .foreign_key(
        'interfaces',
        ('f_target', 'f_tflags', 'f_RTA_OIF'),
        ('f_target', 'f_tflags', 'f_index'),
    )
    .foreign_key(
        'interfaces',
        ('f_target', 'f_tflags', 'f_RTA_IIF'),
        ('f_target', 'f_tflags', 'f_index'),
    )
)

nh_schema = (
    nh.sql_schema()
    .push('route_id', 'TEXT')
    .push('nh_id', 'INTEGER')
    .unique_index('route_id', 'nh_id')
    .foreign_key('routes', ('f_route_id',), ('f_route_id',))
    .foreign_key(
        'interfaces',
        ('f_target', 'f_tflags', 'f_oif'),
        ('f_target', 'f_tflags', 'f_index'),
    )
)

metrics_schema = (
    rtmsg.metrics.sql_schema()
    .push('route_id', 'TEXT')
    .unique_index('route_id')
    .foreign_key('routes', ('f_route_id',), ('f_route_id',))
)

mpls_enc_schema = (
    rtmsg.mpls_encap_info.sql_schema()
    .push('route_id', 'TEXT')
    .unique_index('route_id')
    .foreign_key('routes', ('f_route_id',), ('f_route_id',))
)

init = {
    'specs': [
        ['routes', rt_schema],
        ['nh', nh_schema],
        ['metrics', metrics_schema],
        ['enc_mpls', mpls_enc_schema],
    ],
    'classes': [
        ['routes', rtmsg],
        ['nh', nh],
        ['metrics', rtmsg.metrics],
        ['enc_mpls', rtmsg.mpls_encap_info],
    ],
    'event_map': {rtmsg: [load_rtmsg]},
}


class Via(OrderedDict):
    def __init__(self, prime=None):
        super(OrderedDict, self).__init__()
        if prime is None:
            prime = {}
        elif not isinstance(prime, dict):
            raise TypeError()
        self['family'] = prime.get('family', AF_INET)
        self['addr'] = prime.get('addr', '0.0.0.0')

    def __eq__(self, right):
        return (
            isinstance(right, (dict))
            and self['family'] == right.get('family', AF_INET)
            and self['addr'] == right.get('addr', '0.0.0.0')
        )

    def __repr__(self):
        return repr(dict(self))


class Route(RTNL_Object):
    table = 'routes'
    msg_class = rtmsg
    hidden_fields = ['route_id']
    api = 'route'
    field_filter = RouteFieldFilter

    _replace_on_key_change = True

    @classmethod
    def _count(cls, view):
        if view.chain:
            return view.ndb.task_manager.db_fetchone(
                'SELECT count(*) FROM %s WHERE f_RTA_OIF = %s'
                % (view.table, view.ndb.schema.plch),
                [view.chain['index']],
            )
        else:
            return view.ndb.task_manager.db_fetchone(
                'SELECT count(*) FROM %s' % view.table
            )

    @classmethod
    def _dump_where(cls, view):
        if view.chain:
            plch = view.ndb.schema.plch
            where = '''
                    WHERE
                        main.f_target = %s AND
                        main.f_RTA_OIF = %s
                    ''' % (
                plch,
                plch,
            )
            values = [view.chain['target'], view.chain['index']]
        else:
            where = ''
            values = []
        return (where, values)

    @classmethod
    def summary(cls, view):
        req = '''
              WITH main AS
                  (SELECT
                      nr.f_target, nr.f_tflags, nr.f_RTA_TABLE,
                      nr.f_RTA_DST, nr.f_dst_len,
                      CASE WHEN nh.f_oif > nr.f_RTA_OIF
                          THEN nh.f_oif
                          ELSE nr.f_RTA_OIF
                      END AS f_RTA_OIF,
                      CASE WHEN nh.f_RTA_GATEWAY IS NOT NULL
                          THEN nh.f_RTA_GATEWAY
                          ELSE nr.f_RTA_GATEWAY
                      END AS f_RTA_GATEWAY
                   FROM
                       routes AS nr
                   LEFT JOIN nh
                   ON
                       nr.f_route_id = nh.f_route_id AND
                       nr.f_target = nh.f_target)
              SELECT
                  main.f_target, main.f_tflags, main.f_RTA_TABLE,
                  intf.f_IFLA_IFNAME, main.f_RTA_DST, main.f_dst_len,
                  main.f_RTA_GATEWAY
              FROM
                  main
              INNER JOIN interfaces AS intf
              ON
                  main.f_rta_oif = intf.f_index AND
                  main.f_target = intf.f_target
              '''
        yield (
            'target',
            'tflags',
            'table',
            'ifname',
            'dst',
            'dst_len',
            'gateway',
        )
        where, values = cls._dump_where(view)
        for record in view.ndb.task_manager.db_fetch(req + where, values):
            yield record

    @classmethod
    def dump(cls, view):
        req = '''
              SELECT main.f_target,main.f_tflags,%s
              FROM routes AS main
              LEFT JOIN nh AS nh
              ON main.f_route_id = nh.f_route_id
                  AND main.f_target = nh.f_target
              ''' % ','.join(
            ['%s' % x for x in _dump_rt + _dump_nh + ['main.f_route_id']]
        )
        header = (
            ['target', 'tflags']
            + [rtmsg.nla2name(x[7:]) for x in _dump_rt]
            + ['nh_%s' % nh.nla2name(x[5:]) for x in _dump_nh]
            + ['metrics', 'encap']
        )
        yield header
        plch = view.ndb.schema.plch
        where, values = cls._dump_where(view)
        for record in view.ndb.task_manager.db_fetch(req + where, values):
            route_id = record[-1]
            record = list(record[:-1])
            if route_id is not None:
                #
                # fetch metrics
                metrics = tuple(
                    view.ndb.task_manager.db_fetch(
                        '''
                    SELECT * FROM metrics WHERE f_route_id = %s
                '''
                        % (plch,),
                        (route_id,),
                    )
                )
                if metrics:
                    ret = {}
                    names = view.ndb.schema.compiled['metrics']['norm_names']
                    for k, v in zip(names, metrics[0]):
                        if v is not None and k not in (
                            'target',
                            'route_id',
                            'tflags',
                        ):
                            ret[k] = v
                    record.append(json.dumps(ret))
                else:
                    record.append(None)
                #
                # fetch encap
                enc_mpls = tuple(
                    view.ndb.task_manager.db_fetch(
                        '''
                    SELECT * FROM enc_mpls WHERE f_route_id = %s
                '''
                        % (plch,),
                        (route_id,),
                    )
                )
                if enc_mpls:
                    record.append(enc_mpls[0][2])
                else:
                    record.append(None)
            else:
                record.extend((None, None))
            yield record

    @classmethod
    def spec_normalize(cls, processed, spec):
        if isinstance(spec, basestring):
            processed['dst'] = spec
        return processed

    @classmethod
    def compare_record(self, left, right):
        if isinstance(right, str):
            return right == f'{left["dst"]}/{left["dst_len"]}'

    def _cmp_target(key, self, right):
        right = [MPLSTarget(x) for x in json.loads(right)]
        return all([x[0] == x[1] for x in zip(self[key], right)])

    def _cmp_via(self, right):
        return self['via'] == Via(json.loads(right))

    def _cmp_encap(self, right):
        return all([x[0] == x[1] for x in zip(self.get('encap', []), right)])

    fields_cmp = {
        'dst': partial(_cmp_target, 'dst'),
        'src': partial(_cmp_target, 'src'),
        'newdst': partial(_cmp_target, 'newdst'),
        'encap': _cmp_encap,
        'via': _cmp_via,
    }

    def mark_tflags(self, mark):
        plch = (self.schema.plch,) * 4
        self.schema.execute(
            '''
                            UPDATE interfaces SET
                                f_tflags = %s
                            WHERE
                                (f_index = %s OR f_index = %s)
                                AND f_target = %s
                            '''
            % plch,
            (mark, self['iif'], self['oif'], self['target']),
        )

    def __init__(self, *argv, **kwarg):
        kwarg['iclass'] = rtmsg
        self.event_map = {rtmsg: "load_rtnlmsg"}
        dict.__setitem__(self, 'multipath', [])
        dict.__setitem__(self, 'metrics', MetricsStub(self))
        dict.__setitem__(self, 'deps', 0)
        super(Route, self).__init__(*argv, **kwarg)

    def complete_key(self, key):
        ret_key = {}
        if isinstance(key, basestring):
            ret_key['dst'] = key
        elif isinstance(key, (Record, tuple, list)):
            return super(Route, self).complete_key(key)
        elif isinstance(key, dict):
            ret_key.update(key)
        else:
            raise TypeError('unsupported key type')

        if 'target' not in ret_key:
            ret_key['target'] = self.ndb.localhost

        ##
        # previously here was a code that injected the default
        # table == 254 into the key:
        #
        # table = ret_key.get('table', ret_key.get('RTA_TABLE', 254))
        # if 'table' not in ret_key:
        #     ret_key['table'] = table
        #
        # the issue with the code is that self.exists() didn't use
        # it, thus it was possible to get self.exists() == True and
        # at the same time loading from the DB resulted in an empty
        # record
        #
        # probably more correct behaviour would be to raise KeyError
        # if a route spec has no table defined, and the route is
        # in another table than 254; but for now routes['ipaddr/mask']
        # returns records even outside of the main table

        if isinstance(ret_key.get('dst_len'), basestring):
            ret_key['dst_len'] = int(ret_key['dst_len'])

        if isinstance(ret_key.get('dst'), basestring):
            if ret_key.get('dst') == 'default':
                ret_key['dst'] = ''
                ret_key['dst_len'] = 0
            elif '/' in ret_key['dst']:
                ret_key['dst'], ret_key['dst_len'] = ret_key['dst'].split('/')

        if ret_key.get('family', 0) == AF_MPLS:
            for field in ('dst', 'src', 'newdst', 'via'):
                value = ret_key.get(field, key.get(field, None))
                if isinstance(value, (list, tuple, dict)):
                    ret_key[field] = json.dumps(value)

        return super(Route, self).complete_key(ret_key)

    @property
    def clean(self):
        clean = True
        for s in (self['metrics'],) + tuple(self['multipath']):
            if hasattr(s, 'changed'):
                clean &= len(s.changed) == 0
        return clean & super(Route, self).clean

    def make_req(self, prime):
        req = dict(prime)
        for key in self.changed:
            req[key] = self[key]
        if self['multipath']:
            req['multipath'] = self['multipath']
        if self['metrics']:
            req['metrics'] = self['metrics']
        if self.get('encap') and self.get('encap_type'):
            req['encap'] = {
                'type': self['encap_type'],
                'labels': self['encap'],
            }
        if self.get('gateway'):
            req['gateway'] = self['gateway']
        return req

    @check_auth('obj:modify')
    def __setitem__(self, key, value):
        if key == 'route_id':
            raise ValueError('route_id is read only')
        elif key == 'multipath':
            super(Route, self).__setitem__('multipath', [])
            for mp in value:
                mp = dict(mp)
                if self.state == 'invalid':
                    mp['create'] = True
                obj = NextHop(
                    self, self.view, mp, auth_managers=self.auth_managers
                )
                obj.state.set(self.state.get())
                self['multipath'].append(obj)
            if key in self.changed:
                self.changed.remove(key)
        elif key == 'metrics':
            value = dict(value)
            if not isinstance(self['metrics'], Metrics):
                value['create'] = True
            obj = Metrics(
                self, self.view, value, auth_managers=self.auth_managers
            )
            obj.state.set(self.state.get())
            super(Route, self).__setitem__('metrics', obj)
            if key in self.changed:
                self.changed.remove(key)
        elif self.get('family', 0) == AF_MPLS and key in (
            'dst',
            'src',
            'newdst',
        ):
            if isinstance(value, (dict, int)):
                value = [value]
            na = []
            target = None
            for label in value:
                target = MPLSTarget(label)
                target['bos'] = 0
                na.append(target)
            target['bos'] = 1
            super(Route, self).__setitem__(key, na)
        else:
            super(Route, self).__setitem__(key, value)

    @check_auth('obj:modify')
    def apply(self, rollback=False, req_filter=None, mode='apply'):
        if (
            (self.get('table') == 255)
            and (self.get('family') == 10)
            and (self.get('proto') == 2)
        ):
            # skip automatic ipv6 routes with proto kernel
            return self
        else:
            if self.get('family', AF_INET) == AF_MPLS and not self.get('dst'):
                dict.__setitem__(self, 'dst', [MPLSTarget()])
            return super(Route, self).apply(rollback, req_filter, mode)

    def load_sql(self, *argv, **kwarg):
        super(Route, self).load_sql(*argv, **kwarg)
        # transform MPLS
        if self.get('family', 0) == AF_MPLS:
            for field in ('newdst', 'dst', 'src', 'via'):
                value = self.get(field, None)
                if isinstance(value, basestring) and value != '':
                    if field == 'via':
                        na = json.loads(value)
                    else:
                        na = [MPLSTarget(x) for x in json.loads(value)]
                    dict.__setitem__(self, field, na)
        #
        # fetch encap deps
        if self['deps'] & F_RTA_ENCAP:
            for _ in range(5):
                enc = tuple(
                    self.task_manager.db_fetch(
                        'SELECT * FROM enc_mpls WHERE f_route_id = %s'
                        % (self.schema.plch,),
                        (self['route_id'],),
                    )
                )
                if enc:
                    na = [MPLSTarget(x) for x in json.loads(enc[0][2])]
                    self.load_value('encap', na)
                    break
                time.sleep(0.1)
            else:
                self.log.error('no encap loaded for %s' % (self['route_id'],))
        #
        #
        if not self.load_event.is_set():
            return
        #
        # fetch metrics
        if self['deps'] & F_RTA_METRICS:
            for _ in range(5):
                metrics = tuple(
                    self.task_manager.db_fetch(
                        'SELECT * FROM metrics WHERE f_route_id = %s'
                        % (self.schema.plch,),
                        (self['route_id'],),
                    )
                )
                if metrics:
                    self['metrics'] = Metrics(
                        self,
                        self.view,
                        {'route_id': self['route_id']},
                        auth_managers=self.auth_managers,
                    )
                    break
                time.sleep(0.1)
            else:
                self.log.error(
                    'no metrics loaded for %s' % (self['route_id'],)
                )
        #
        # fetch multipath
        #
        # FIXME: use self['deps']
        if 'nh_id' not in self and self.get('route_id') is not None:
            nhs = self.task_manager.db_fetch(
                'SELECT * FROM nh WHERE f_route_id = %s' % (self.schema.plch,),
                (self['route_id'],),
            )

            flush = False
            idx = 0
            for nexthop in tuple(self['multipath']):
                if not isinstance(nexthop, NextHop):
                    flush = True

                if not flush:
                    try:
                        spec = next(nhs)
                    except StopIteration:
                        flush = True
                    for key, value in zip(nexthop.names, spec):
                        if key in nexthop and value is None:
                            continue
                        else:
                            nexthop.load_value(key, value)
                if flush:
                    self['multipath'].pop(idx)
                    continue
                idx += 1

            for nexthop in nhs:
                key = {'route_id': self['route_id'], 'nh_id': nexthop[-1]}
                (
                    self['multipath'].append(
                        NextHop(
                            self,
                            self.view,
                            key,
                            auth_managers=self.auth_managers,
                        )
                    )
                )


class RouteSub:
    def apply(self, rollback=False, req_filter=None, mode='apply'):
        return self.route.apply(rollback, req_filter, mode)

    def commit(self):
        return self.route.commit()

    def set(self, key, value):
        self[key] = value

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.commit()


class NextHop(RouteSub, RTNL_Object):
    msg_class = nh
    table = 'nh'
    hidden_fields = ('route_id', 'target')

    def mark_tflags(self, mark):
        plch = (self.schema.plch,) * 4
        self.schema.execute(
            '''
                            UPDATE interfaces SET
                                f_tflags = %s
                            WHERE
                                (f_index = %s OR f_index = %s)
                                AND f_target = %s
                            '''
            % plch,
            (mark, self.route['iif'], self.route['oif'], self.route['target']),
        )

    def __init__(self, route, *argv, **kwarg):
        self.route = route
        kwarg['iclass'] = nh
        kwarg['check'] = False
        super(NextHop, self).__init__(*argv, **kwarg)


class MetricsStub(RouteSub, dict):
    def __init__(self, route):
        self.route = route

    def __setitem__(self, key, value):
        # This assignment forces the Metrics object to replace
        # MetricsStub; it is the MetricsStub object end of life
        self.route['metrics'] = {key: value}

    def __getitem__(self, key):
        raise KeyError('metrics not initialized for this route')


class Metrics(RouteSub, RTNL_Object):
    msg_class = rtmsg.metrics
    table = 'metrics'
    hidden_fields = ('route_id', 'target')

    def mark_tflags(self, mark):
        plch = (self.schema.plch,) * 4
        self.schema.execute(
            '''
                            UPDATE interfaces SET
                                f_tflags = %s
                            WHERE
                                (f_index = %s OR f_index = %s)
                                AND f_target = %s
                            '''
            % plch,
            (mark, self.route['iif'], self.route['oif'], self.route['target']),
        )

    def __init__(self, route, *argv, **kwarg):
        self.route = route
        kwarg['iclass'] = rtmsg.metrics
        kwarg['check'] = False
        super(Metrics, self).__init__(*argv, **kwarg)



"""
Module for viewing and modifying sysctl parameters
"""

import os

import salt.utils.files
from salt.exceptions import CommandExecutionError

# Define the module's virtual name
__virtualname__ = "sysctl"


def __virtual__():
    """
    Only run on Darwin (macOS) systems
    """
    if __grains__["os"] == "MacOS":
        return __virtualname__
    return (
        False,
        "The darwin_sysctl execution module cannot be loaded: "
        "Only available on macOS systems.",
    )


def show(config_file=False):
    """
    Return a list of sysctl parameters for this minion

    config: Pull the data from the system configuration file
        instead of the live data.

    CLI Example:

    .. code-block:: bash

        salt '*' sysctl.show
    """
    roots = (
        "audit",
        "debug",
        "hw",
        "hw",
        "kern",
        "machdep",
        "net",
        "net",
        "security",
        "user",
        "vfs",
        "vm",
    )
    cmd = "sysctl -a"
    ret = {}
    out = __salt__["cmd.run"](cmd, output_loglevel="trace", python_shell=False)
    comps = [""]
    for line in out.splitlines():
        # This might need to be converted to a regex, and more, as sysctl output
        # can for some reason contain entries such as:
        #
        # user.tzname_max = 255
        # kern.clockrate: hz = 100, tick = 10000, profhz = 100, stathz = 100
        # kern.clockrate: {hz = 100, tick = 10000, tickadj = 2, profhz = 100,
        #                 stathz = 100}
        #
        # Yes. That's two `kern.clockrate`.
        #
        if any([line.startswith(f"{root}.") for root in roots]):
            comps = line.split(": " if ": " in line else " = ", 1)
            if len(comps) == 2:
                ret[comps[0]] = comps[1]
            else:
                ret[comps[0]] = ""
        elif comps[0]:
            ret[comps[0]] += f"{line}\n"
        else:
            continue
    return ret


def get(name):
    """
    Return a single sysctl parameter for this minion

    name
        The name of the sysctl value to display.

    CLI Example:

    .. code-block:: bash

        salt '*' sysctl.get hw.physmem
    """
    cmd = f"sysctl -n {name}"
    out = __salt__["cmd.run"](cmd, python_shell=False)
    return out


def assign(name, value):
    """
    Assign a single sysctl parameter for this minion

    name
        The name of the sysctl value to edit.

    value
        The sysctl value to apply.

    CLI Example:

    .. code-block:: bash

        salt '*' sysctl.assign net.inet.icmp.icmplim 50
    """
    ret = {}
    cmd = f'sysctl -w {name}="{value}"'
    data = __salt__["cmd.run_all"](cmd, python_shell=False)

    if data["retcode"] != 0:
        raise CommandExecutionError("sysctl failed: {}".format(data["stderr"]))
    new_name, new_value = data["stdout"].split(":", 1)
    ret[new_name] = new_value.split(" -> ")[-1]
    return ret


def persist(name, value, config="/etc/sysctl.conf", apply_change=False):
    """
    Assign and persist a simple sysctl parameter for this minion

    name
        The name of the sysctl value to edit.

    value
        The sysctl value to apply.

    config
        The location of the sysctl configuration file.

    apply_change
        Default is False; Default behavior only creates or edits
        the sysctl.conf file. If apply is set to True, the changes are
        applied to the system.

    CLI Example:

    .. code-block:: bash

        salt '*' sysctl.persist net.inet.icmp.icmplim 50
        salt '*' sysctl.persist coretemp_load NO config=/etc/sysctl.conf
    """
    nlines = []
    edited = False
    value = str(value)

    # If the sysctl.conf is not present, add it
    if not os.path.isfile(config):
        try:
            with salt.utils.files.fopen(config, "w+") as _fh:
                _fh.write("#\n# Kernel sysctl configuration\n#\n")
        except OSError:
            msg = "Could not write to file: {0}"
            raise CommandExecutionError(msg.format(config))

    with salt.utils.files.fopen(config, "r") as ifile:
        for line in ifile:
            line = salt.utils.stringutils.to_unicode(line)
            if not line.startswith(f"{name}="):
                nlines.append(line)
                continue
            else:
                key, rest = line.split("=", 1)
                if rest.startswith('"'):
                    _, rest_v, rest = rest.split('"', 2)
                elif rest.startswith("'"):
                    _, rest_v, rest = rest.split("'", 2)
                else:
                    rest_v = rest.split()[0]
                    rest = rest[len(rest_v) :]
                if rest_v == value:
                    return "Already set"
                nlines.append(f"{name}={value}\n")
                edited = True
    if not edited:
        nlines.append(f"{name}={value}\n")
    nlines = [salt.utils.stringutils.to_str(_l) for _l in nlines]
    with salt.utils.files.fopen(config, "w+") as ofile:
        ofile.writelines(nlines)
    # If apply_change=True, apply edits to system
    if apply_change is True:
        assign(name, value)
        return "Updated and applied"
    return "Updated"



"""
This module allows you to install certificates into the windows certificate
manager.

.. code-block:: bash

    salt '*' certutil.add_store salt://cert.cer 'TrustedPublisher'
"""

import logging
import os
import re

import salt.utils.platform
from salt.exceptions import CommandExecutionError

log = logging.getLogger(__name__)
__virtualname__ = "certutil"


def __virtual__():
    """
    Only works on Windows
    """
    if salt.utils.platform.is_windows():
        return __virtualname__
    return False, "Module win_certutil: module only works on Windows systems."


def get_cert_serial(cert_file, saltenv="base"):
    """
    Get the serial number of a certificate file

    Args:

        cert_file (str):
            The certificate file to find the serial for. Can be a local file or
            a file on the file server (``salt://``)

        saltenv (:obj:`str`, optional):
            The environment on the fileserver to use for the certificate.
            Default is ``base``.

    Returns:
        str: The serial number of the certificate if found, otherwise ``None``.

    CLI Example:

    .. code-block:: bash

        salt '*' certutil.get_cert_serial <certificate name>
    """
    cert_file = __salt__["cp.cache_file"](cert_file, saltenv)

    # Since we're allowing a path, let's make sure it exists
    if not os.path.exists(cert_file):
        msg = f"cert_file not found: {cert_file}"
        raise CommandExecutionError(msg)

    cmd = f'certutil.exe -silent -verify "{cert_file}"'
    out = __salt__["cmd.run"](cmd)
    # match serial number by paragraph to work with multiple languages
    matches = re.search(r":\s*(\w*)\r\n\r\n", out)
    if matches is not None:
        return matches.groups()[0].strip()
    else:
        return None


def get_stored_cert_serials(store):
    """
    Get all of the certificate serials in the specified store

    store (str):
        The store to get all the certificate serials from

    Returns:
        list: A list of serial numbers found, or an empty list if none found

    CLI Example:

    .. code-block:: bash

        salt '*' certutil.get_stored_cert_serials <store>
    """
    cmd = f'certutil.exe -store "{store}"'
    out = __salt__["cmd.run"](cmd)
    # match serial numbers by header position to work with multiple languages
    matches = re.findall(r"={16}\r\n.*:\s*(\w*)\r\n", out)
    return matches


def add_store(source, store, retcode=False, saltenv="base"):
    """
    Add the cert to the given Certificate Store

    Args:

        source (str):
            The source certificate file. This is either the path to a local file or
            a file from the file server in the form of ``salt://path/to/file``

        store (str):
            The certificate store to add the certificate to

        retcode (:obj:`bool`, optional):
            If ``True``, return the retcode instead of stdout.
            Default is ``False``.

        saltenv (:obj:`str`, optional):
            The salt environment to use. This is ignored if the path is local.
            Default is ``base``.

    CLI Example:

    .. code-block:: bash

        salt '*' certutil.add_store salt://cert.cer TrustedPublisher
        salt '*' certutil.add_store C:\\path\\to\\local.cer TrustedPublisher
    """
    source = __salt__["cp.cache_file"](source, saltenv)

    # Since we're allowing a path, let's make sure it exists
    if not os.path.exists(source):
        msg = f"cert_file not found: {source}"
        raise CommandExecutionError(msg)

    cmd = f'certutil.exe -addstore {store} "{source}"'
    if retcode:
        return __salt__["cmd.retcode"](cmd)
    else:
        return __salt__["cmd.run"](cmd)


def del_store(source, store, retcode=False, saltenv="base"):
    """
    Delete the cert from the given Certificate Store

    Args:

        source (str):
            The source certificate file. This is either the path to a local file or
            a file from the file server in the form of ``salt://path/to/file``

        store (str):
            The certificate store to delete the certificate from

        retcode (:obj:`bool`, optional):
            If ``True``, return the retcode instead of stdout.
            Default is ``False``

        saltenv (:obj:`str`, optional):
            The salt environment to use. This is ignored if the path is local.
            Default is ``base``.

    CLI Example:

    .. code-block:: bash

        salt '*' certutil.del_store salt://cert.cer TrustedPublisher
        salt '*' certutil.del_store 'C:\\path\\to\\local.cer' TrustedPublisher
    """
    source = __salt__["cp.cache_file"](source, saltenv)

    # Since we're allowing a path, let's make sure it exists
    if not os.path.exists(source):
        msg = f"cert_file not found: {source}"
        raise CommandExecutionError(msg)

    serial = get_cert_serial(source)
    cmd = f'certutil.exe -delstore {store} "{serial}"'
    if retcode:
        return __salt__["cmd.retcode"](cmd)
    else:
        return __salt__["cmd.run"](cmd)



"""Functions for Retrieving data from graphite"""
from typing import Union, List, Optional, Tuple
from datetime import datetime
import urllib.parse
import requests
from arrow.arrow import Arrow
import arrow

QUERY_FMT = '%H:%M_%Y%m%d'


class GraphiteError(Exception):
    """Raised on error when fetching graphite data"""


class Graphite:
    """Performs Graphite Queries"""

    def __init__(
        self,
        render_url: Optional[str] = 'https://graphite.imhadmin.net/render/',
        timeout: Optional[Union[float, Tuple[float]]] = 600,
    ):
        self.render_url = render_url
        self.timeout = timeout

    @staticmethod
    def _params(
        target: Union[str, List[str]],
        start: Optional[Union[datetime, Arrow, str, int]] = None,
        stop: Optional[Union[datetime, Arrow, str, int]] = None,
        tzname: Optional[str] = None,
    ) -> dict:
        # int -> arrow
        if isinstance(start, int):
            start = arrow.get(start)
        if isinstance(stop, int):
            stop = arrow.get(stop)
        # format datetimes
        if isinstance(start, (datetime, Arrow)):
            if tzname is None:  # if tzname was not overridden
                tzname = start.tzname()
            start = start.strftime(QUERY_FMT)
        if isinstance(stop, (datetime, Arrow)):
            stop = stop.strftime(QUERY_FMT)
        # graphite's pytz may be old
        if tzname in ('PDT', 'PST'):
            tzname = 'America/Los_Angeles'
        if tzname in ('EST', 'EDT'):
            tzname = 'America/New_York'
        params = {'target': target}
        if start is not None:
            params['from'] = start
        if stop is not None:
            params['until'] = stop
        if tzname is not None:
            params['tz'] = tzname
        return params

    def query(
        self,
        target: Union[str, List[str]],
        start: Optional[Union[datetime, Arrow, str, int]] = None,
        stop: Optional[Union[datetime, Arrow, str, int]] = None,
        tzname: Optional[str] = None,
    ) -> list:
        """Query data from graphite

        Args:
            target (str or list): graphite bucket(s) to fetch data for
            start: start time to fetch data for. If not supplied, it'll
                be 24 hours ago **(see note)**
            stop: stop time to fetch data for. If not supplied, it'll be now
                **(see note)**
            tzname: timezone to use. If not supplied, the ``start`` arg's
                timezone will be used. If the ``start`` arg is not timezone
                aware, it'll use graphite defaults

        Raises:
            GraphiteError: on any error

        Note:
            ``start`` and ``stop`` will accept ``arrow`` objects, ``datetimes``,
            ``int`` for unix timestamps, and ``str``. If it's a ``str``, it'll
            be sent as a literal and handled per graphite's documentation
            https://graphite.readthedocs.io/en/latest/render_api.html
        """
        params = self._params(target, start, stop, tzname)
        params['format'] = 'json'
        try:
            req = requests.get(
                self.render_url,
                timeout=self.timeout,
                params=params,
            )
        except requests.RequestException as exc:
            raise GraphiteError(exc) from exc
        try:
            return req.json()
        except ValueError as exc:
            if req.status_code != 200:
                raise GraphiteError(
                    f"{req.url}\nHTTP {req.status_code}: {req.reason}"
                ) from exc
            raise GraphiteError(
                f"{req.url}\n{type(exc).__name__}: {exc}"
            ) from exc

    def query_dict(
        self,
        target: Union[str, List[str]],
        start: Optional[Union[datetime, Arrow, str, int]] = None,
        stop: Optional[Union[datetime, Arrow, str, int]] = None,
        tzname: Optional[str] = None,
    ) -> dict:
        """Just like ``query`` except it converts the data to a dict

        Args:
            target (str or list): graphite bucket(s) to fetch data for
            start: start time to fetch data for. If not supplied, it'll
                be 24 hours ago **(see note)**
            stop: stop time to fetch data for. If not supplied, it'll be now
                **(see note)**
            tzname: timezone to use. If not supplied, the ``start`` arg's
                timezone will be used. If the ``start`` arg is not timezone
                aware, it'll use graphite defaults

        Raises:
            GraphiteError: on any error

        Note:
            ``start`` and ``stop`` will accept ``arrow`` objects, ``datetimes``,
            ``int`` for unix timestamps, and ``str``. If it's a ``str``, it'll
            be sent as a literal and handled per graphite's documentation
            https://graphite.readthedocs.io/en/latest/render_api.html
        """
        return {
            x['target']: x['datapoints']
            for x in self.query(target, start, stop, tzname)
        }

    def img_url(
        self,
        target: Union[str, List[str]],
        start: Optional[Union[datetime, Arrow, str, int]] = None,
        stop: Optional[Union[datetime, Arrow, str, int]] = None,
        tzname: Optional[str] = None,
        **kwargs,
    ) -> str:
        """Gets a graphite render URL for an image but does not fetch it

        Args:
            target (str or list): graphite bucket(s) to fetch data for
            start: start time to fetch data for. If not supplied, it'll
                be 24 hours ago **(see note)**
            stop: stop time to fetch data for. If not supplied, it'll be now
                **(see note)**
            tzname: timezone to use. If not supplied, the ``start`` arg's
                timezone will be used. If the ``start`` arg is not timezone
                aware, it'll use graphite defaults

        Raises:
            GraphiteError: on any error

        Note:
            ``start`` and ``stop`` will accept ``arrow`` objects, ``datetimes``,
            ``int`` for unix timestamps, and ``str``. If it's a ``str``, it'll
            be sent as a literal and handled per graphite's documentation
            https://graphite.readthedocs.io/en/latest/render_api.html
        """
        params = self._params(target, start, stop, tzname)
        params.update(kwargs)
        args = urllib.parse.urlencode(query=params, doseq=True)
        return f'{self.render_url}?{args}'



"""
Encapsulate the different transports available to Salt.
"""

import logging
import warnings

from salt.transport.base import (
    TRANSPORTS,
    ipc_publish_client,
    ipc_publish_server,
    publish_client,
    publish_server,
    request_client,
    request_server,
)

log = logging.getLogger(__name__)

# Suppress warnings when running with a very old pyzmq. This can be removed
# after we drop support for Ubuntu 16.04 and Debian 9
warnings.filterwarnings(
    "ignore", message="IOLoop.current expected instance.*", category=RuntimeWarning
)

__all__ = (
    "TRANSPORTS",
    "ipc_publish_client",
    "ipc_publish_server",
    "publish_client",
    "publish_server",
    "request_client",
    "request_server",
)



import typing as t

if t.TYPE_CHECKING:
    from .runtime import Undefined


class TemplateError(Exception):
    """Baseclass for all template errors."""

    def __init__(self, message: t.Optional[str] = None) -> None:
        super().__init__(message)

    @property
    def message(self) -> t.Optional[str]:
        return self.args[0] if self.args else None


class TemplateNotFound(IOError, LookupError, TemplateError):
    """Raised if a template does not exist.

    .. versionchanged:: 2.11
        If the given name is :class:`Undefined` and no message was
        provided, an :exc:`UndefinedError` is raised.
    """

    # Silence the Python warning about message being deprecated since
    # it's not valid here.
    message: t.Optional[str] = None

    def __init__(
        self,
        name: t.Optional[t.Union[str, "Undefined"]],
        message: t.Optional[str] = None,
    ) -> None:
        IOError.__init__(self, name)

        if message is None:
            from .runtime import Undefined

            if isinstance(name, Undefined):
                name._fail_with_undefined_error()

            message = name

        self.message = message
        self.name = name
        self.templates = [name]

    def __str__(self) -> str:
        return str(self.message)


class TemplatesNotFound(TemplateNotFound):
    """Like :class:`TemplateNotFound` but raised if multiple templates
    are selected.  This is a subclass of :class:`TemplateNotFound`
    exception, so just catching the base exception will catch both.

    .. versionchanged:: 2.11
        If a name in the list of names is :class:`Undefined`, a message
        about it being undefined is shown rather than the empty string.

    .. versionadded:: 2.2
    """

    def __init__(
        self,
        names: t.Sequence[t.Union[str, "Undefined"]] = (),
        message: t.Optional[str] = None,
    ) -> None:
        if message is None:
            from .runtime import Undefined

            parts = []

            for name in names:
                if isinstance(name, Undefined):
                    parts.append(name._undefined_message)
                else:
                    parts.append(name)

            parts_str = ", ".join(map(str, parts))
            message = f"none of the templates given were found: {parts_str}"

        super().__init__(names[-1] if names else None, message)
        self.templates = list(names)


class TemplateSyntaxError(TemplateError):
    """Raised to tell the user that there is a problem with the template."""

    def __init__(
        self,
        message: str,
        lineno: int,
        name: t.Optional[str] = None,
        filename: t.Optional[str] = None,
    ) -> None:
        super().__init__(message)
        self.lineno = lineno
        self.name = name
        self.filename = filename
        self.source: t.Optional[str] = None

        # this is set to True if the debug.translate_syntax_error
        # function translated the syntax error into a new traceback
        self.translated = False

    def __str__(self) -> str:
        # for translated errors we only return the message
        if self.translated:
            return t.cast(str, self.message)

        # otherwise attach some stuff
        location = f"line {self.lineno}"
        name = self.filename or self.name
        if name:
            location = f'File "{name}", {location}'
        lines = [t.cast(str, self.message), "  " + location]

        # if the source is set, add the line to the output
        if self.source is not None:
            try:
                line = self.source.splitlines()[self.lineno - 1]
            except IndexError:
                pass
            else:
                lines.append("    " + line.strip())

        return "\n".join(lines)

    def __reduce__(self):  # type: ignore
        # https://bugs.python.org/issue1692335 Exceptions that take
        # multiple required arguments have problems with pickling.
        # Without this, raises TypeError: __init__() missing 1 required
        # positional argument: 'lineno'
        return self.__class__, (self.message, self.lineno, self.name, self.filename)


class TemplateAssertionError(TemplateSyntaxError):
    """Like a template syntax error, but covers cases where something in the
    template caused an error at compile time that wasn't necessarily caused
    by a syntax error.  However it's a direct subclass of
    :exc:`TemplateSyntaxError` and has the same attributes.
    """


class TemplateRuntimeError(TemplateError):
    """A generic runtime error in the template engine.  Under some situations
    Jinja may raise this exception.
    """


class UndefinedError(TemplateRuntimeError):
    """Raised if a template tries to operate on :class:`Undefined`."""


class SecurityError(TemplateRuntimeError):
    """Raised if a template tries to do something insecure if the
    sandbox is enabled.
    """


class FilterArgumentError(TemplateRuntimeError):
    """This error is raised if a filter was called with inappropriate
    arguments
    """



"""
    :codeauthor: Pedro Algarvio (pedro@algarvio.me)


    salt.utils.immutabletypes
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    Immutable types
"""

import copy
from collections.abc import Mapping, Sequence, Set


class ImmutableDict(Mapping):
    """
    An immutable dictionary implementation
    """

    def __init__(self, obj):
        self.__obj = obj

    def __len__(self):
        return len(self.__obj)

    def __iter__(self):
        return iter(self.__obj)

    def __getitem__(self, key):
        return freeze(self.__obj[key])

    def __repr__(self):
        return f"<{self.__class__.__name__} {repr(self.__obj)}>"

    def __deepcopy__(self, memo):
        return copy.deepcopy(self.__obj)

    def copy(self):
        """
        Return an un-frozen copy of self
        """
        return copy.deepcopy(self.__obj)


class ImmutableList(Sequence):
    """
    An immutable list implementation
    """

    def __init__(self, obj):
        self.__obj = obj

    def __len__(self):
        return len(self.__obj)

    def __iter__(self):
        return iter(self.__obj)

    def __add__(self, other):
        return self.__obj + other

    def __radd__(self, other):
        return other + self.__obj

    def __getitem__(self, key):
        return freeze(self.__obj[key])

    def __repr__(self):
        return f"<{self.__class__.__name__} {repr(self.__obj)}>"

    def __deepcopy__(self, memo):
        return copy.deepcopy(self.__obj)

    def copy(self):
        """
        Return an un-frozen copy of self
        """
        return copy.deepcopy(self.__obj)


class ImmutableSet(Set):
    """
    An immutable set implementation
    """

    def __init__(self, obj):
        self.__obj = obj

    def __len__(self):
        return len(self.__obj)

    def __iter__(self):
        return iter(self.__obj)

    def __contains__(self, key):
        return key in self.__obj

    def __repr__(self):
        return f"<{self.__class__.__name__} {repr(self.__obj)}>"

    def __deepcopy__(self, memo):
        return copy.deepcopy(self.__obj)

    def copy(self):
        """
        Return an un-frozen copy of self
        """
        return copy.deepcopy(self.__obj)


def freeze(obj):
    """
    Freeze python types by turning them into immutable structures.
    """
    if isinstance(obj, dict):
        return ImmutableDict(obj)
    if isinstance(obj, list):
        return ImmutableList(obj)
    if isinstance(obj, set):
        return ImmutableSet(obj)
    return obj



"""Code for accessing the backup authority server api

.. data:: bakauth.AUTH_JSON:

    "/opt/backups/etc/auth.json"
    *(str)*

.. data:: bakauth.BAKAUTH1:

    "ash-sys-pro-bakauth1.imhadmin.net"
    *(str)*

.. data:: bakauth.BAKAUTH2:

    "ash-sys-dev-bakauth2.imhadmin.net"
    *(str)*

.. data:: bakauth.BAKAUTH3:

    "lax-sys-pro-bakauth3.imhadmin.net"
    *(str)*

.. data:: bakauth.SHARED_CLASSES:

    {'imh_reseller', 'imh_shared', 'hub_shared', 'imh_ngx'}
    *(set)*
"""

import sys
from typing import Union, Any
import functools
import logging
import random
import json
import time
import platform
import distro
import rads
from restic import Restic, ResticRepo
from .hints import (
    SharedFailoverLocks,
    AgentClientLookup,
    AgentCpuserLookup,
    VznodeBackupLookup,
    VznodeRestoreLookup,
    UserBuckets,
    RegDetails,
)
from .sess import Status, MdsState, post, DEFAULT_TIMEOUT, DEFAULT_RETRIES
from .exc import (
    BakAuthError,
    BakAuthDown,
    ClusterUnavailable,
    AMPDownError,
    WrongServerClass,
    BakAuthLoginFailed,
    BakAuthWrongLogin,
    VpsRestricted,
    LookupMissing,
    WrongSharedServer,
    NoAmpAccount,
    DedicatedMoved,
    InternalQuota,
    Unregistered,
)

__version__ = '1.5.5'


BAKAUTH1 = 'ash-sys-pro-bakauth1.imhadmin.net'  # prod main
BAKAUTH2 = 'ash-sys-dev-bakauth2.imhadmin.net'  # testing
BAKAUTH3 = 'lax-sys-pro-bakauth3.imhadmin.net'  # prod replicant

AUTH_JSON = '/opt/backups/etc/auth.json'

SHARED_CLASSES = {'imh_reseller', 'imh_shared', 'hub_shared', 'imh_ngx'}


class BakAuth:
    """Handles backup authority requests"""

    def __init__(self):
        try:
            with open(AUTH_JSON, encoding='utf-8') as handle:
                data = json.load(handle)
        except FileNotFoundError as exc:
            raise Unregistered(str(exc)) from exc
        self._post = functools.partial(
            post, auth=(data['apiuser'], data['authkey'])
        )

    def _post_main(
        self,
        *,
        uri: str,
        timeout: int,
        retries: int,
        log_retries: bool = True,
        **data,
    ) -> tuple[Status, Any]:
        """Perform a post request that only the primary bakauth server can
        handle

        Args:
            uri (str): HTTP request URI
            timeout (int): HTTP request timeout in seconds
            retries (int): HTTP request auto-retries after timeout
            log_retries (bool): whether to log on auto-retries. Defaults False
            **data: POST form data

        Returns:
            tuple[Status, Any]: (``Status`` enum, data)
        """
        return self._post(
            bakauth_host=BAKAUTH1,
            uri=uri,
            timeout=timeout,
            retries=retries,
            log_retries=log_retries,
            data=data,
        )

    def _post_pref(
        self,
        *,
        uri: str,
        timeout: int,
        retries: int,
        log_retries: bool = True,
        pref_main: bool = True,
        **data,
    ) -> tuple[Status, Any, str]:
        """Perform a post request that the primary bakauth server should handle,
        but will failover to a replicant bakauth server if needed

        Args:
            uri (str): HTTP request URI
            timeout (int): HTTP request timeout in seconds
            retries (int): HTTP request auto-retries after timeout
            log_retries (bool): whether to log on auto-retries. Defaults False
            pref_main (bool): If true, try bakauth1 first. If false,
                try bakauth3 first. Defaults True.
            **data: POST form data

        Returns:
            tuple[Status, Any, str]: (``Status`` enum, data, bakauth host used)
        """
        kwargs = {
            'uri': uri,
            'timeout': timeout,
            'retries': retries,
            'log_retries': log_retries,
            'data': data,
        }
        if pref_main:
            bakauth_hosts = [BAKAUTH1, BAKAUTH3]
        else:
            bakauth_hosts = [BAKAUTH3, BAKAUTH1]
        status, data = self._post(bakauth_host=bakauth_hosts[0], **kwargs)
        if status is Status.REQUEST_EXCEPTION:
            if log_retries:
                logging.warning(
                    '%s::%s: request exception - retrying using %s',
                    bakauth_hosts[0],
                    uri,
                    bakauth_hosts[1],
                )
            ret = self._post(bakauth_host=bakauth_hosts[1], **kwargs)
            return *ret, bakauth_hosts[1]
        return status, data, bakauth_hosts[0]

    def _post_either(
        self,
        *,
        uri: str,
        timeout: int,
        retries: int,
        log_retries: bool = True,
        **data,
    ) -> tuple[Status, Any]:
        """Perform a post request that any production bakauth server can
        handle (round robin)

        Args:
            uri (str): HTTP request URI
            timeout (int): HTTP request timeout in seconds
            retries (int): HTTP request auto-retries after timeout
            log_retries (bool): whether to log on auto-retries. Defaults False
            **data: POST form data

        Returns:
            tuple[Status, Any]: (``Status`` enum, data)
        """
        prio = [BAKAUTH1, BAKAUTH3]
        random.shuffle(prio)
        kwargs = {
            'uri': uri,
            'timeout': timeout,
            'retries': retries,
            'log_retries': log_retries,
            'data': data,
        }
        status, data = self._post(bakauth_host=prio[0], **kwargs)
        if status is Status.REQUEST_EXCEPTION:
            if log_retries:
                logging.warning(
                    '%s:%s: request exception - retrying using %s',
                    prio[0],
                    uri,
                    prio[1],
                )
            return self._post(bakauth_host=prio[1], **kwargs)
        return status, data

    def task_wait(
        self,
        task_id: str,
        *,
        wait_mins: int = 240,
        poll_secs: int = 3,
        bakauth_host: str = BAKAUTH1,
        timeout: int = DEFAULT_TIMEOUT,
        retries: int = DEFAULT_TIMEOUT,
    ) -> None:
        """Wait for a celery task to finish on the Backup Authority server

        Args:
            task_id (str): task identifier string
            wait_mins (int, optional): max minutes to wait. Defaults to 240.
            poll_secs (int, optional): secondss between checking the state of
                the task. Defaults to 3.
            timeout (int, optional): request timeout in secs
                (per poll request, not in total)
            retries (int): number of times to retry the request on timeout

        Raises:
            BakAuthError: error checking the state of the task
        """
        state_msg = 'QUEUED'
        wait = wait_mins * 60
        start = time.time()
        while state_msg in ('QUEUED', 'STARTED') and time.time() - start < wait:
            time.sleep(poll_secs)
            status, state_msg = self._post(
                bakauth_host=bakauth_host,
                uri='/lookup/check_task',
                timeout=timeout,
                retries=retries,
                log_retries=True,
                data={'task_id': task_id},
            )
            if status is not Status.OKAY:
                raise BakAuthError(status=status, data=state_msg)

    @staticmethod
    def register(
        *,
        svr_class: str,
        host: str,
        require_amp: bool = False,
    ) -> dict[str, str]:
        """Register a (non-internal) server in backup authority

        Args:
            svr_class (str): server classification, one in:
                imh_vps, imh_reseller, imh_shared, hub_shared, imh_ngx,
                or imh_ded
            host (str): short hostname which should match what AMP knows
            require_amp (bool): if True, instruct bakauth to reject registration
                if the supplied host does not match a known AMP account

        Raises:
            BakAuthError: any error registering the server

        Returns:
            dict[str, str]: contains keys "authkey" and "apiuser"
        """
        status, data = post(
            bakauth_host=BAKAUTH1,
            uri='/register',
            timeout=DEFAULT_TIMEOUT,
            retries=DEFAULT_RETRIES,
            log_retries=True,
            data={
                'host': host,
                'svr_class': svr_class,
                'require_amp': '1' if require_amp else '0',
            },
        )
        if status is not Status.OKAY:
            raise BakAuthError(status=status, data=data)
        return data

    def vzclient_backup(
        self,
        *,
        veids: dict[int, str],
        net: Union[str, None] = None,
        timeout: int = DEFAULT_TIMEOUT,
        retries: int = DEFAULT_RETRIES,
    ) -> VznodeBackupLookup:
        """Requests information form bakauth needed to backup a vz node

        Args:
            veids (dict[int, str]): If performing a vps backup run, provide the
                IDs of all vps found on this compute node, mapped to their
                FQDNs. If only backing up mds, send an empty dict
            net (str | None): set "lan" or "wan" to override which
                ceph network to use
            timeout (int): request timeout in seconds
            retries (int): number of times to retry the request on timeout

        Raises:
            BakAuthError: error looking up data

        Returns:
            VznodeBackupLookup: restic info needed to backup a vznode.
            Contains keys "endpoints", "node_keys", and "vps_keys"
        """
        status, data = self._post_main(
            uri='/vzclient/backup',
            veids=json.dumps(veids),
            net=net,
            timeout=timeout,
            retries=retries,
        )
        if status is not Status.OKAY:
            raise BakAuthError(status=status, data=data)
        for task_id in data['task_ids']:
            try:
                self.task_wait(task_id, timeout=timeout, retries=retries)
            except BakAuthError as exc:
                logging.warning('%s', exc)
        return {
            'endpoints': data['endpoints'],
            'node_keys': ResticRepo(**data['node_keys']),
            'vps_keys': {
                int(k): ResticRepo(**v) for k, v in data['vps_keys'].items()
            },
            'changed': {int(k): v for k, v in data['changed'].items()},
        }

    def vzclient_restore(
        self,
        veid: Union[int, None],
        *,
        net: Union[str, None] = None,
        timeout: int = DEFAULT_TIMEOUT,
        retries: int = DEFAULT_RETRIES,
    ) -> VznodeRestoreLookup:
        """Requests information from bakauth needed to restore data for a vps
        from a vznode

        Args:
            veid (int | None): ID for the vps you're trying to restore.
                Explicitly set to None to fetch info for the node itself.
            net (str | None): set "lan" or "wan" to override which
                ceph network to use
            timeout (int): request timeout in seconds
            retries (int): number of times to retry the request on timeout

        Raises:
            BakAuthDown: if bakauth could not be reached
            BakAuthLoginFailed: API user/password is wrong
            LookupMissing: if a requested vps's keys do not exist
            VpsRestricted: if a requested vps is internal
            BakAuthError: catch-all for any other api error

        Returns:
            VznodeRestoreLookup: restic info needed to restore a vps from a
            vznode. The dict has keys "this_endpoint", "key_info",
            and "all_endpoints"
        """
        if veid is None:
            kwargs = {'node': '1'}
        else:
            kwargs = {'veid': veid}
        status, data = self._post_either(
            uri='/vzclient/restore',
            net=net,
            timeout=timeout,
            retries=retries,
            **kwargs,
        )
        if status is not Status.OKAY:
            raise BakAuthError(status=status, data=data)
        data['key_info'] = ResticRepo(**data['key_info'])
        return data

    def get_shared_failover_locks(
        self,
        user: str,
        timeout: int = DEFAULT_TIMEOUT,
        retries: int = DEFAULT_RETRIES,
    ) -> SharedFailoverLocks:
        """Check which failover backups should not be rotated"""
        status, data, _ = self._post_pref(
            uri='/failover/get_shared_locks',
            timeout=timeout,
            retries=retries,
            pref_main=False,
            user=user,
        )
        if status is not Status.OKAY:
            raise BakAuthError(status=status, data=data)
        return data

    def monitoring_vz_update(
        self,
        *,
        version: str,
        mds_state: MdsState,
        vps_crit: int,
        vps_warn: int,
        vps_sizes: dict[int, int],
        timeout: int = DEFAULT_TIMEOUT,
        retries: int = DEFAULT_RETRIES,
    ) -> None:
        """Post backup monitoring status for a vznode

        Args:
            version (str): RPM version
            mds_state (MdsState): MDS status enum
            vps_crit (int): number of vps old enough to be a nrpe critical
            vps_warn (int): number of vps behind schedule but not critical
            vps_sizes (dict[int, int]): veids mapped to VPS sizes in MiB
            timeout (int): request timeout in seconds
            retries (int): number of times to retry the request on timeout

        Raises:
            BakAuthError: error updating monitoring status
        """
        py_version = sys.version_info
        status, data = self._post_main(
            uri='/monitoring/vz_update',
            version=version,
            mds_state=mds_state.value,
            vps_crit=vps_crit,
            vps_warn=vps_warn,
            vps_sizes=json.dumps(vps_sizes),
            os_info=json.dumps(distro.info()),
            py_info=f'{py_version[0]}.{py_version[1]}',
            timeout=timeout,
            retries=retries,
        )
        if status is not Status.OKAY:
            raise BakAuthError(status=status, data=data)

    def monitoring_update(
        self,
        *,
        version: str,
        running: bool,
        ded_moved: bool,
        errors: list[str],
        num_old: int,
        timeout: int = DEFAULT_TIMEOUT,
        retries: int = DEFAULT_RETRIES,
    ) -> None:
        """Updates bakauth with client server status

        Args:
            version (str): backup client RPM version
            running (bool): whether the backup runner daemon is running
            ded_moved (bool): True if this is a dedi and its bucket was moved
                to a new server
            errors (list[str]): error messages to display in monitoring dash
            num_old (int): number of old tasks in queue
            timeout (int): request timeout in seconds
            retries (int): number of times to retry the request on timeout

        Raises:
            BakAuthError: error updating monitoring status
        """
        py_version = sys.version_info
        status, data = self._post_main(
            uri='/monitoring/update',
            timeout=timeout,
            retries=retries,
            version=version,
            running=1 if running else 0,
            ded_moved=1 if ded_moved else 0,
            errors=json.dumps(errors),
            num_old=num_old,
            os_info=json.dumps(distro.info()),
            py_info=f'{py_version[0]}.{py_version[1]}',
        )
        if status is not Status.OKAY:
            raise BakAuthError(status=status, data=data)

    def note_auto_ticket(
        self,
        *,
        task: str,
        plugin: str,
        user: str,
        ipaddr: str,
        timeout: int = DEFAULT_TIMEOUT,
        retries: int = DEFAULT_RETRIES,
    ) -> None:
        """Add an account note when a ticket is generated

        Args:
            task (str): backup/restore task name
            plugin (str): plugin task name (e.g. "cPanel Backup Manager")
            user (str): username
            ipaddr (str): IP address of user
            timeout (int): request timeout in seconds
            retries (int): number of times to retry the request on timeout

        Raises:
            BakAuthError: error posting note
        """
        status, data = self._post_main(
            uri='/note/auto_ticket',
            timeout=timeout,
            retries=retries,
            task=task,
            plugin=plugin,
            user=user,
            ipaddr=ipaddr,
        )
        if status is not Status.OKAY:
            raise BakAuthError(status=status, data=data)

    def post_user_sizes(
        self,
        user_sizes: dict[str, dict[str, int]],
        *,
        notify: list[str],
        reset: list[str],
        timeout: int = DEFAULT_TIMEOUT,
        retries: int = DEFAULT_RETRIES,
    ) -> None:
        """Post a user's account size for AMP to access

        Args:
            user_sizes (dict[str, dict[str, int]]): user account sizes (in MiB).
                each dict should have keys "total_mb" and "usage_mb"
            notify (list[str]): users over quota
            reset (list[str]): users under quota (to reset their notify times)
            timeout (int): request timeout in seconds
            retries (int): number of times to retry the request on timeout

        Raises:
            BakAuthError: any error updating usage
        """
        status, data = self._post_main(
            uri='/usage/set/shared',
            timeout=timeout,
            retries=retries,
            users=json.dumps(user_sizes),
            notify=json.dumps(notify),
            reset=json.dumps(reset),
        )
        if status is not Status.OKAY:
            raise BakAuthError(status=status, data=data)

    def post_vded_size(
        self,
        usage: int,
        total: int,
        *,
        notify: bool = False,
        reset: bool = False,
        timeout: int = DEFAULT_TIMEOUT,
        retries: int = DEFAULT_RETRIES,
    ) -> None:
        """Post a v/ded server's size for AMP to access

        Args:
            usage (int): disk usage of selected backups (in MiB)
            total (int): disk usage of all account data (in MiB)
            notify (bool, optional): notify as over quota. Defaults to False.
            reset (bool, optional): reset notify counter. Defaults to False.
            timeout (int): request timeout in seconds
            retries (int): number of times to retry the request on timeout

        Raises:
            BakAuthError: any error updating usage
        """
        status, data = self._post_main(
            uri='/usage/set/vded',
            timeout=timeout,
            retries=retries,
            usage=usage,
            total=total,
            notify='1' if notify else '0',
            reset='1' if reset else '0',
        )
        if status is not Status.OKAY:
            raise BakAuthError(status=status, data=data)

    def agent_get_user_bucket(
        self,
        user: str,
        key: str,
        timeout: int = DEFAULT_TIMEOUT,
        retries: int = DEFAULT_RETRIES,
    ) -> AgentCpuserLookup:
        """Used by support tooling to request a shared user's bucket details
        from a server other than where the bucket belongs

        Args:
            user (str): username
            key (str): agent key from cpjump
            timeout (int): request timeout in seconds
            retries (int): number of times to retry the request on timeout

        Raises:
            BakAuthError: any error requesting bucket details

        Returns:
            AgentCpuserLookup: restic auth info and wans available
        """
        status, data = self._post_main(
            uri='/agent/get_bucket/cpuser',
            timeout=timeout,
            retries=retries,
            cpuser=user,
            key=key,
        )
        if status is not Status.OKAY:
            raise BakAuthError(status=status, data=data)
        wans = data.pop('wans')
        return {'wans': wans, 'repo': ResticRepo(**data)}

    def agent_get_server_bucket(
        self,
        host: str,
        key: str,
        timeout: int = DEFAULT_TIMEOUT,
        retries: int = DEFAULT_RETRIES,
    ) -> AgentClientLookup:
        """Used by support tooling to request v/ded bucket details for a
        server with an active registration, from a server other than where
        the bucket belongs

        Args:
            host (str): short hostname
            key (str): agent key from cpjump
            timeout (int): request timeout in seconds
            retries (int): number of times to retry the request on timeout

        Raises:
            BakAuthError: any error requesting bucket details

        Returns:
            AgentClientLookup: restic auth info and wans available
        """
        status, data = self._post_main(
            uri='/agent/get_bucket/client',
            timeout=timeout,
            retries=retries,
            host=host,
            key=key,
        )
        if status is not Status.OKAY:
            raise BakAuthError(status=status, data=data)
        wans = data.pop('wans')
        svr_class = data.pop('svr_class')
        return {
            'wans': wans,
            'svr_class': svr_class,
            'repo': ResticRepo(**data),
        }

    def agent_get_stashed_bucket(
        self,
        bucket: str,
        key: str,
        timeout: int = DEFAULT_TIMEOUT,
        retries: int = DEFAULT_RETRIES,
    ) -> AgentClientLookup:
        """Used by support tooling (baksync) to request v/ded bucket details
        for a bucket which is marked for deletion

        Args:
            bucket (str): full bucket name from cpjump
            key (str): agent key from cpjump
            timeout (int): request timeout in seconds
            retries (int): number of times to retry the request on timeout

        Raises:
            BakAuthError: any error requesting bucket details

        Returns:
            AgentClientLookup: restic auth info and wans available
        """
        status, data = self._post_main(
            uri='/agent/get_bucket/client',
            timeout=timeout,
            retries=retries,
            stashed_bucket=bucket,
            key=key,
        )
        if status is not Status.OKAY:
            raise BakAuthError(status=status, data=data)
        wans = data.pop('wans')
        svr_class = data.pop('svr_class')
        return {
            'wans': wans,
            'svr_class': svr_class,
            'repo': ResticRepo(**data),
        }

    def get_user_buckets(
        self,
        users: list[str],
        *,
        wait_mins: int,
        nocache: bool = False,
        suspends: Union[dict, None] = None,
        timeout: int = DEFAULT_TIMEOUT,
        retries: int = DEFAULT_RETRIES,
    ) -> UserBuckets:
        """Get quota and bucket details for multiple shared users

        Args:
            users (list[str]): users to get bucket data for
            wait_mins (int): max number of minutes to wait for bucket creation
                (per cluster)
            nocache (bool): Instruct bakauth to skip cache when
                looking up quota information. Defaults to False.
            suspends (dict | None): only used by the scheduler
                cron. This is a dict of suspended users and when/why.
            timeout (int): request timeout in seconds
            retries (int): number of times to retry the request on timeout

        Raises:
            BakAuthError: any error requesting bucket details

        Returns:
            UserBuckets: dict where "quotas_gb" contains a dict
            of usernames to quotas in GiB. "repos" contains a dict of usernames
            to ResticRepo objects. "missing" is specific to the replicant,
            ``BAKAUTH3``. If ``BAKAUTH1`` did not reply, we use
            ``BAKAUTH3``, which is unable to setup new users. "copy_users" are
            users which get cross-coast backups
        """
        if not isinstance(users, list):
            raise TypeError
        kwargs = {'users': json.dumps(users), 'nocache': 1 if nocache else 0}
        if suspends is not None:
            kwargs['suspends'] = json.dumps(suspends)
        status, data, _ = self._post_pref(
            uri='/buckets/users',
            timeout=timeout,
            retries=retries,
            **kwargs,
        )
        if status is not Status.OKAY:
            raise BakAuthError(status=status, data=data)
        if wait_mins:
            for task_id in data['task_ids']:
                try:
                    self.task_wait(task_id, wait_mins=wait_mins)
                except BakAuthError as exc:
                    logging.warning(
                        'error when waiting for ceph user setup: %s', exc
                    )
        quotas = {k: v.pop('quota') for k, v in data['users'].items()}
        return {
            'copy_users': data['copy_users'],
            'quotas_gb': quotas,
            'repos': {k: ResticRepo(**v) for k, v in data['users'].items()},
            'missing': data['missing'],
        }

    def get_user_bucket(
        self,
        user: str,
        *,
        timeout: int = DEFAULT_TIMEOUT,
        retries: int = DEFAULT_RETRIES,
    ) -> tuple[int, ResticRepo]:
        """Like get_user_buckets but for only one user and wait_mins=0.

        Args:
            user (str): username
            timeout (int): request timeout in seconds
            retries (int): number of times to retry the request on timeout

        Raises:
            WrongSharedServer: The requested user does not belong here
            BakAuthDown: Backup Authority server was unreachable
            BakAuthError: any other error

        Returns:
            tuple[int, ResticRepo]: quota in GiB and ResticRepo object
        """
        if not isinstance(user, str):
            raise TypeError(f'get_user_bucket({user=})')
        # BakAuthError may raise here
        lookup = self.get_user_buckets(
            [user], nocache=True, wait_mins=0, timeout=timeout, retries=retries
        )
        if user not in lookup['repos']:
            raise WrongSharedServer(
                status=Status.ERROR,
                data=f'{user} is not assigned to this server',
            )
        if user in lookup['missing']:
            # This means we tried bakauth1 and failed, then successfully
            # reached bakauth2, which replied saying the bucket was
            # missing, which is something only bakauth1 can fix.
            raise BakAuthDown(
                status=Status.REQUEST_EXCEPTION,
                data=f'Could not connect to {BAKAUTH1}',
            )
        return lookup['quotas_gb'][user], lookup['repos'][user]

    def get_user_bucket_v2(
        self,
        user: str,
        *,
        timeout: int = DEFAULT_TIMEOUT,
        retries: int = DEFAULT_RETRIES,
    ) -> tuple[int, ResticRepo, bool]:
        """Like get_user_buckets but for only one user and wait_mins=0.

        Args:
            user (str): username
            timeout (int): request timeout in seconds
            retries (int): number of times to retry the request on timeout

        Raises:
            WrongSharedServer: The requested user does not belong here
            BakAuthDown: Backup Authority server was unreachable
            BakAuthError: any other error

        Returns:
            tuple[int, ResticRepo, bool]: quota in GiB, ResticRepo object, and
                whether this user gets cross-coast backups.
        """
        if not isinstance(user, str):
            raise TypeError(f'get_user_bucket({user=})')
        # BakAuthError may raise here
        lookup = self.get_user_buckets(
            [user], nocache=True, wait_mins=0, timeout=timeout, retries=retries
        )
        if user not in lookup['repos']:
            raise WrongSharedServer(
                status=Status.ERROR,
                data=f'{user} is not assigned to this server',
            )
        if user in lookup['missing']:
            # This means we tried bakauth1 and failed, then successfully
            # reached bakauth2, which replied saying the bucket was
            # missing, which is something only bakauth1 can fix.
            raise BakAuthDown(
                status=Status.REQUEST_EXCEPTION,
                data=f'Could not connect to {BAKAUTH1}',
            )
        geo = user in lookup['copy_users']
        return lookup['quotas_gb'][user], lookup['repos'][user], geo

    def get_failover_limit(
        self,
        *,
        timeout: int = DEFAULT_TIMEOUT,
        retries: int = DEFAULT_RETRIES,
    ) -> int:
        """Get the shared failover account size limit in GiB

        Args:
            timeout (int): request timeout in seconds
            retries (int): number of times to retry the request on timeout

        Raises:
            BakAuthError: any error requesting the size limit

        Returns:
            int: shared failover account size limit in GiB
        """
        status, data = self._post_either(
            uri='/buckets/failover_limit',
            timeout=timeout,
            retries=retries,
        )
        if status is not Status.OKAY:
            raise BakAuthError(status=status, data=data)
        return data['failover_gib']

    def get_vded_quota(
        self,
        *,
        nocache=False,
        timeout=DEFAULT_TIMEOUT,
        retries=DEFAULT_RETRIES,
    ) -> int:
        """Get this v/ded server's quota as an int in GiB

        Args:
            nocache (bool): Instruct bakauth to skip cache when
                looking up quota information. Defaults to False.
            timeout (int): request timeout in seconds
            retries (int): number of times to retry the request on timeout

        Raises:
            BakAuthError: any error requesting the server's quota

        Returns:
            int: this v/ded server's quota as an int in GiB
        """
        status, data = self._post_either(
            uri='/buckets/vded_quota',
            timeout=timeout,
            retries=retries,
            ver=1,
            nocache=int(nocache),
        )
        if status is not Status.OKAY:
            raise BakAuthError(status=status, data=data)
        return data

    def get_vded_quota_v2(
        self,
        *,
        nocache=False,
        timeout=DEFAULT_TIMEOUT,
        retries=DEFAULT_RETRIES,
    ) -> tuple[int, bool]:
        """Get this v/ded server's quota as an int in GiB

        Args:
            nocache (bool): Instruct bakauth to skip cache when
                looking up quota information. Defaults to False.
            timeout (int): request timeout in seconds
            retries (int): number of times to retry the request on timeout

        Raises:
            BakAuthError: any error requesting the server's quota

        Returns:
            int: this v/ded server's quota as an int in GiB
            bool: whether this server gets cross-coast backups
        """
        status, data = self._post_either(
            uri='/buckets/vded_quota',
            timeout=timeout,
            retries=retries,
            ver=2,
            nocache=int(nocache),
        )
        if status is not Status.OKAY:
            raise BakAuthError(status=status, data=data)
        return data['quota'], data['copy']

    def get_reg_details(
        self,
        *,
        net: Union[str, None] = None,
        startup: bool = False,
        timeout: int = DEFAULT_TIMEOUT,
        retries: int = DEFAULT_RETRIES,
    ) -> RegDetails:
        """Fetch registration details needed for backup-runner to start

        Args:
            startup (bool): If True, bakauth performs extra checks on buckets.
                This should only be set to True on initial backup-runner
                startup. Defaults false.
            net(str, optional): "wan" or "lan" to request that ceph endpoint
            timeout (int): request timeout in seconds
            retries (int): number of times to retry the request on timeout

        Raises:
            BakAuthLoginFailed: auth.json is invalid
            BakAuthWrongLogin: this server has the wrong server's auth.json
            BakAuthError: any other error getting the server's registration info

        Returns:
            RegDetails: contains "svr_class", "client_host",
            "endpoint", and "repo"
        """
        data = {}
        if net:
            data['net'] = net
        if startup:
            data['startup'] = '1'
        status, data = self._post_either(
            uri='/buckets/reg_details',
            timeout=timeout,
            retries=retries,
            **data,
        )
        if status is not Status.OKAY:
            raise BakAuthError(status=status, data=data)
        if data['svr_class'] in SHARED_CLASSES:
            if data['client_host'] != platform.node().split('.')[0]:
                raise BakAuthWrongLogin(
                    status=Status.ERROR,
                    data="Shared server registration incorrect! "
                    f"Registered as {data['client_host']!r}",
                )

        return {
            'name': data['name'],
            'svr_class': data['svr_class'],
            'client_host': data['client_host'],
            'endpoint': data['endpoint'],
            'location': data['location'],
            'wans': data['wans'],
            'copy': data['copy'],
            'repo': ResticRepo(
                bucket=data['bucket'],
                restic_pass=data['restic_pass'],
                access_key=data['access_key'],
                secret_key=data['secret_key'],
            ),
        }

    def ping(
        self,
        timeout: int = DEFAULT_TIMEOUT,
        retries: int = DEFAULT_RETRIES,
        log_retries: bool = False,
    ) -> None:
        """Check if backup authority is reachable and raise an exception if not

        Args:
            timeout (int): request timeout in seconds
            retries (int): number of times to retry each server on timeout
            log_retries (bool): whether to log on auto-retries. Defaults False

        Raises:
            BakAuthError: backup authority is not reachable
        """
        status, data = self._post_either(
            uri='/monitoring/ping',
            timeout=timeout,
            retries=retries,
            log_retries=log_retries,
        )
        if status is not Status.OKAY:
            raise BakAuthError(status=status, data=data)

    def check_bucket_index(
        self,
        *,
        user: str,
        bucket: str,
        timeout: int = DEFAULT_TIMEOUT,
        retries: int = DEFAULT_RETRIES,
        geo: Union[bool, None] = None,
    ) -> None:
        """Tell ceph to repair a bucket's index

        Args:
            user (str): username that owns the bucket (root if v/ded)
            bucket (str): full bucket name
            timeout (int): request timeout in seconds
            retries (int): number of times to retry the request on timeout
            geo (bool | None): if True, check opposite coast; if False, check
                same-coast. If None (default), check both if present.

        Raises:
            BakAuthError: error either requesting the repair, or checking on it
        """
        if geo is None:
            kwargs = {}
        else:
            kwargs = {'geo': int(geo)}
        # queue the repair
        status, data, auth_host = self._post_pref(
            uri='/buckets/queue_repair',
            timeout=timeout,
            retries=retries,
            pref_main=False,
            user=user,
            bucket=bucket,
            **kwargs,
        )
        if status is not Status.OKAY:
            raise BakAuthError(status=status, data=data)
        # wait until it's finished
        # use wait_mins=16 because celery crits after 15 mins
        # task_wait can also raise BakAuthError
        self.task_wait(data, wait_mins=16, poll_secs=10, bakauth_host=auth_host)

    def get_restic(
        self, user: str = 'root', *, geo: bool = False, **kwargs
    ) -> Restic:
        """Get a Restic instance for a specific user on a backups 3.x client,
        assuming the current server is setup with it

        Args:
            user (str): If on a shared server, this is the username to get the
                Restic instance for. For reseller children, supply their
                reseller's name. If on vps/dedicated, always lookup "root"
            geo (bool): If True, use the secondary, geographically separated
                cluster.
            **kwargs: other keyword arguments to send as-is to the
                Restic() constructor

        Raises:
            ValueError: Incorrect user argument supplied
            WrongSharedServer: The requested user does not belong here
            BakAuthDown: Backup Authority server was unreachable
            BakAuthError: Any other API error

        Returns:
            Restic: restic instance
        """
        reg = self.get_reg_details()
        if geo:
            if not all(
                (
                    reg['copy']['endpoint'],
                    reg['copy']['name'],
                    reg['copy']['location'],
                )
            ):
                raise ClusterUnavailable(
                    status=Status.ERROR, data="This cluster is not available"
                )
            endpoint = reg['copy']['endpoint']
            cluster = reg['copy']['name']
        else:
            endpoint = reg['endpoint']
            cluster = reg['name']
        is_shared = reg['svr_class'] in SHARED_CLASSES
        if not is_shared and user != 'root':
            raise ValueError(f'{user=}; should be root on {reg["svr_class"]}')
        if user == 'root':
            return Restic(
                endpoint=endpoint, cluster=cluster, repo=reg['repo'], **kwargs
            )
        _, repo = self.get_user_bucket(user)
        return Restic(endpoint=endpoint, cluster=cluster, repo=repo, **kwargs)

    def all_restics(
        self, users: Union[list[str], None] = None, geo: bool = False, **kwargs
    ) -> dict[str, Restic]:
        """If users are not supplied, get all restic instances for this shared
        server, including root. Otherwise, get that set of users

        Args:
            users (list[str] | None): List of users to obtain restic
                instances for. This function will not resolve child account
                names to resellers, so be sure to only request main cpanel users
            geo (bool): If True, use the secondary, geographically separated
                cluster.
            **kwargs: other keyword arguments to send as-is to the
                Restic() constructor

        Raises:
            BakAuthDown: Backup Authority server was unreachable
            WrongServerClass: The current server is not a shared server
            BakAuthError: Any other API error

        Returns:
            dict[str, Restic]: usernames mapped to Restic instances. If users
                was supplied, not all usernames requested may be in the result,
                if bakauth did not recognize some as belonging to this server
        """
        if users is None:
            users = rads.main_cpusers()
        repos = self.get_user_buckets(users=users, wait_mins=0)['repos']
        reg = self.get_reg_details()
        if geo:
            if not all(
                reg['copy']['endpoint'],
                reg['copy']['name'],
                reg['copy']['location'],
            ):
                raise ClusterUnavailable(
                    status=Status.ERROR, data="This cluster is not available"
                )
            endpoint = reg['copy']['endpoint']
            cluster = reg['copy']['name']
        else:
            endpoint = reg['endpoint']
            cluster = reg['name']
        ret = {
            'root': Restic(
                endpoint=endpoint, cluster=cluster, repo=reg['repo'], **kwargs
            )
        }
        for user, repo in repos.items():
            ret[user] = Restic(
                endpoint=endpoint, cluster=cluster, repo=repo, **kwargs
            )
        return ret



'''
General request and RTNL object data filters.
'''

import weakref
from collections import ChainMap


class RequestProcessor(dict):
    def __init__(self, field_filter=None, context=None, prime=None):
        self.field_filter = field_filter
        self.context = (
            context if isinstance(context, (dict, weakref.ProxyType)) else {}
        )
        self.combined = ChainMap(self, self.context)
        if isinstance(prime, dict):
            self.update(prime)

    def __setitem__(self, key, value):
        if value is None:
            return
        if key in self:
            del self[key]
        for nkey, nvalue in self.filter(key, value).items():
            super(RequestProcessor, self).__setitem__(nkey, nvalue)

    def filter(self, key, value):
        if hasattr(self.field_filter, '_key_transform'):
            key = self.field_filter._key_transform(key)
        if (
            hasattr(self.field_filter, '_allowed')
            and key not in self.field_filter._allowed
        ):
            return {}
        if hasattr(
            self.field_filter, 'policy'
        ) and not self.field_filter.policy(key):
            return {}
        return getattr(
            self.field_filter, f'set_{key}', lambda *argv: {key: value}
        )(self.combined, value)

    def update(self, prime):
        for key, value in tuple(prime.items()):
            self[key] = value

    def set_filter(self, field_filter):
        self.field_filter = field_filter
        return self

    def apply_filter(self, field_filter):
        self.field_filter = field_filter
        self.update(self)
        return self

    def finalize(self, cmd_context=None):
        if hasattr(self.field_filter, 'finalize_for_iproute'):
            # old interface
            self.field_filter.finalize_for_iproute(self.combined, cmd_context)
        if hasattr(self.field_filter, 'finalize'):
            # new interface
            self.field_filter.finalize(self.combined)
        return self



"""
Module for making various web calls. Primarily designed for webhooks and the
like, but also useful for basic http testing.

.. versionadded:: 2015.5.0
"""

import logging

import salt.utils.http

log = logging.getLogger(__name__)


def query(url, output=True, **kwargs):
    """
    Query a resource, and decode the return data

    Passes through all the parameters described in the
    :py:func:`utils.http.query function <salt.utils.http.query>`:

    CLI Example:

    .. code-block:: bash

        salt-run http.query http://somelink.com/
        salt-run http.query http://somelink.com/ method=POST \
            params='key1=val1&key2=val2'
        salt-run http.query http://somelink.com/ method=POST \
            data='<xml>somecontent</xml>'
    """
    if output is not True:
        log.warning("Output option has been deprecated. Please use --quiet.")
    if "node" not in kwargs:
        kwargs["node"] = "master"
    opts = __opts__.copy()
    if "opts" in kwargs:
        opts.update(kwargs["opts"])
        del kwargs["opts"]

    ret = salt.utils.http.query(url=url, opts=opts, **kwargs)
    return ret


def update_ca_bundle(target=None, source=None, merge_files=None):
    """
    Update the local CA bundle file from a URL

    .. versionadded:: 2015.5.0

    CLI Example:

    .. code-block:: bash

        salt-run http.update_ca_bundle
        salt-run http.update_ca_bundle target=/path/to/cacerts.pem
        salt-run http.update_ca_bundle source=https://example.com/cacerts.pem

    If the ``target`` is not specified, it will be pulled from the ``ca_cert``
    configuration variable available to the master. If it cannot be found there,
    it will be placed at ``<<FILE_ROOTS>>/cacerts.pem``.

    If the ``source`` is not specified, it will be pulled from the
    ``ca_cert_url`` configuration variable available to the master. If it cannot
    be found, it will be downloaded from the cURL website, using an http (not
    https) URL. USING THE DEFAULT URL SHOULD BE AVOIDED!

    ``merge_files`` may also be specified, which includes a string or list of
    strings representing a file or files to be appended to the end of the CA
    bundle, once it is downloaded.

    CLI Example:

    .. code-block:: bash

        salt-run http.update_ca_bundle merge_files=/path/to/mycert.pem
    """
    return salt.utils.http.update_ca_bundle(target, source, __opts__, merge_files)



"""
Beacon to monitor disk usage.

.. versionadded:: 2015.5.0

:depends: python-psutil
"""

import logging
import re

import salt.utils.beacons
import salt.utils.platform

try:
    import psutil

    HAS_PSUTIL = True
except ImportError:
    HAS_PSUTIL = False

log = logging.getLogger(__name__)

__virtualname__ = "diskusage"


def __virtual__():
    if HAS_PSUTIL is False:
        err_msg = "psutil library is missing."
        log.error("Unable to load %s beacon: %s", __virtualname__, err_msg)
        return False, err_msg
    else:
        return __virtualname__


def validate(config):
    """
    Validate the beacon configuration
    """
    # Configuration for diskusage beacon should be a list of dicts
    if not isinstance(config, list):
        return False, "Configuration for diskusage beacon must be a list."
    return True, "Valid beacon configuration"


def beacon(config):
    r"""
    Monitor the disk usage of the minion

    Specify thresholds for each disk and only emit a beacon if any of them are
    exceeded.

    .. code-block:: yaml

        beacons:
          diskusage:
            - /: 63%
            - /mnt/nfs: 50%

    Windows drives must be quoted to avoid yaml syntax errors

    .. code-block:: yaml

        beacons:
          diskusage:
            -  interval: 120
            - 'c:\\': 90%
            - 'd:\\': 50%

    Regular expressions can be used as mount points.

    .. code-block:: yaml

        beacons:
          diskusage:
            - '^\/(?!home).*$': 90%
            - '^[a-zA-Z]:\\$': 50%

    The first one will match all mounted disks beginning with "/", except /home
    The second one will match disks from A:\ to Z:\ on a Windows system

    Note that if a regular expression are evaluated after static mount points,
    which means that if a regular expression matches another defined mount point,
    it will override the previously defined threshold.

    """
    whitelist = []
    config = salt.utils.beacons.remove_hidden_options(config, whitelist)
    parts = psutil.disk_partitions(all=True)
    ret = []
    for mounts in config:
        mount = next(iter(mounts))

        # Because we're using regular expressions
        # if our mount doesn't end with a $, insert one.
        mount_re = mount
        if not mount.endswith("$"):
            mount_re = f"{mount}$"

        if salt.utils.platform.is_windows():
            # mount_re comes in formatted with a $ at the end
            # can be `C:\\$` or `C:\\\\$`
            # re string must be like `C:\\\\` regardless of \\ or \\\\
            # also, psutil returns uppercase
            mount_re = re.sub(r":\\\$", r":\\\\", mount_re)
            mount_re = re.sub(r":\\\\\$", r":\\\\", mount_re)
            mount_re = mount_re.upper()

        for part in parts:
            if re.match(mount_re, part.mountpoint):
                _mount = part.mountpoint

                try:
                    _current_usage = psutil.disk_usage(_mount)
                except OSError:
                    log.warning("%s is not a valid mount point.", _mount)
                    continue

                current_usage = _current_usage.percent
                monitor_usage = mounts[mount]
                if isinstance(monitor_usage, str) and "%" in monitor_usage:
                    monitor_usage = re.sub("%", "", monitor_usage)
                monitor_usage = float(monitor_usage)
                if current_usage >= monitor_usage:
                    ret.append({"diskusage": current_usage, "mount": _mount})
    return ret



"""
Management of Gentoo configuration using eselect
================================================

A state module to manage Gentoo configuration via eselect

"""

# Define a function alias in order not to shadow built-in's
__func_alias__ = {"set_": "set"}


def __virtual__():
    """
    Only load if the eselect module is available in __salt__
    """
    if "eselect.exec_action" in __salt__:
        return "eselect"
    return (False, "eselect module could not be loaded")


def set_(name, target, module_parameter=None, action_parameter=None):
    """
    Verify that the given module is set to the given target

    name
        The name of the module

    target
        The target to be set for this module

    module_parameter
        additional params passed to the defined module

    action_parameter
        additional params passed to the defined action

    .. code-block:: yaml

        profile:
          eselect.set:
            - target: hardened/linux/amd64
    """
    ret = {"changes": {}, "comment": "", "name": name, "result": True}

    old_target = __salt__["eselect.get_current_target"](
        name, module_parameter=module_parameter, action_parameter=action_parameter
    )

    if target == old_target:
        ret["comment"] = "Target '{}' is already set on '{}' module.".format(
            target, name
        )
    elif target not in __salt__["eselect.get_target_list"](
        name, action_parameter=action_parameter
    ):
        ret["comment"] = "Target '{}' is not available for '{}' module.".format(
            target, name
        )
        ret["result"] = False
    elif __opts__["test"]:
        ret["comment"] = f"Target '{target}' will be set on '{name}' module."
        ret["result"] = None
    else:
        result = __salt__["eselect.set_target"](
            name,
            target,
            module_parameter=module_parameter,
            action_parameter=action_parameter,
        )
        if result:
            ret["changes"][name] = {"old": old_target, "new": target}
            ret["comment"] = f"Target '{target}' set on '{name}' module."
        else:
            ret["comment"] = "Target '{}' failed to be set on '{}' module.".format(
                target, name
            )
            ret["result"] = False
    return ret



"""
Functions to work with JSON
"""

import json
import logging

import salt.utils.data
import salt.utils.stringutils

log = logging.getLogger(__name__)


# One to one mappings
JSONEncoder = json.JSONEncoder


def __split(raw):
    """
    Performs a splitlines on the string. This function exists to make mocking
    possible in unit tests, since the member functions of the str/unicode
    builtins cannot be mocked.
    """
    return raw.splitlines()


def find_json(raw):
    """
    Pass in a raw string and load the json when it starts. This allows for a
    string to start or end with garbage but the JSON be cleanly loaded
    """
    ret = {}
    lines = __split(raw)
    lengths = list(map(len, lines))
    starts = []
    ends = []

    # Search for possible starts and ends of the json fragments
    for ind, line in enumerate(lines):
        line = line.lstrip()
        line = line[0] if line else line
        if line == "{" or line == "[":
            starts.append((ind, line))
        if line == "}" or line == "]":
            ends.append((ind, line))

    # List all the possible pairs of starts and ends,
    # and fill the length of each block to sort by size after
    starts_ends = []
    for start, start_char in starts:
        for end, end_br in reversed(ends):
            if end > start and (
                (start_char == "{" and end_br == "}")
                or (start_char == "[" and end_br == "]")
            ):
                starts_ends.append((start, end, sum(lengths[start : end + 1])))

    # Iterate through all the possible pairs starting from the largest
    starts_ends.sort(key=lambda x: (x[2], x[1] - x[0], x[0]), reverse=True)
    for start, end, _ in starts_ends:
        # Try filtering non-JSON text right after the last closing character
        end_str = lines[end].lstrip()[0]
        working = "\n".join(lines[start:end]) + end_str
        try:
            ret = json.loads(working)
            return ret
        except ValueError:
            continue

    # Fall back to old implementation for backward compatibility
    # expecting json after the text
    for ind, _ in enumerate(lines):
        try:
            working = "\n".join(lines[ind:])
        except UnicodeDecodeError:
            working = "\n".join(salt.utils.data.decode(lines[ind:]))

        try:
            ret = json.loads(working)
        except ValueError:
            continue
        if ret:
            return ret
    if not ret:
        # Not json, raise an error
        raise ValueError


def import_json():
    """
    Import a json module, starting with the quick ones and going down the list)
    """
    for fast_json in ("ujson", "yajl", "json"):
        try:
            mod = __import__(fast_json)
            log.trace("loaded %s json lib", fast_json)
            return mod
        except ImportError:
            continue


def load(fp, **kwargs):
    """
    .. versionadded:: 2018.3.0

    Wraps json.load

    You can pass an alternate json module (loaded via import_json() above)
    using the _json_module argument)
    """
    return kwargs.pop("_json_module", json).load(fp, **kwargs)


def loads(s, **kwargs):
    """
    .. versionadded:: 2018.3.0

    Wraps json.loads and prevents a traceback in the event that a bytestring is
    passed to the function. (Python < 3.6 cannot load bytestrings)

    You can pass an alternate json module (loaded via import_json() above)
    using the _json_module argument)
    """
    json_module = kwargs.pop("_json_module", json)
    try:
        return json_module.loads(s, **kwargs)
    except TypeError as exc:
        # json.loads cannot load bytestrings in Python < 3.6
        if isinstance(s, bytes):
            return json_module.loads(salt.utils.stringutils.to_unicode(s), **kwargs)
        else:
            raise


def dump(obj, fp, **kwargs):
    """
    .. versionadded:: 2018.3.0

    Wraps json.dump, and assumes that ensure_ascii is False (unless explicitly
    passed as True) for unicode compatibility. Note that setting it to True
    will mess up any unicode characters, as they will be dumped as the string
    literal version of the unicode code point.

    On Python 2, encodes the result to a str since json.dump does not want
    unicode types.

    You can pass an alternate json module (loaded via import_json() above)
    using the _json_module argument)
    """
    json_module = kwargs.pop("_json_module", json)
    if "ensure_ascii" not in kwargs:
        kwargs["ensure_ascii"] = False
    return json_module.dump(obj, fp, **kwargs)


def dumps(obj, **kwargs):
    """
    .. versionadded:: 2018.3.0

    Wraps json.dumps, and assumes that ensure_ascii is False (unless explicitly
    passed as True) for unicode compatibility. Note that setting it to True
    will mess up any unicode characters, as they will be dumped as the string
    literal version of the unicode code point.

    On Python 2, encodes the result to a str since json.dumps does not want
    unicode types.

    You can pass an alternate json module (loaded via import_json() above)
    using the _json_module argument)
    """
    json_module = kwargs.pop("_json_module", json)
    if "ensure_ascii" not in kwargs:
        kwargs["ensure_ascii"] = False
    return json_module.dumps(obj, **kwargs)



"""Functions for posting data to carbon"""

import socket
import time
import random
from typing import Optional, List

RELAY_PORTS = [2003, 2005, 2007, 2009]


def carbon_post(
    data: List[str],
    server: str = 'graphite.imhadmin.net',
    port: Optional[int] = None,
    chunk_size: int = 500,
    chunk_sleep: float = 0.2,
    max_tries: int = 3,
) -> bool:
    """Post all items in a list of strings to Carbon

    Args:
        data: lines to post, each formatted ``"bucket.name value timestamp"``
        server: graphite server FQDN to post to
        port: TCP port to post to; if unspecified, a random port
            from ``mtrlib.carbon.RELAY_PORTS`` will be used
        chunk_size: how many metrics to post in each HTTP POST request
        chunk_sleep: how long in seconds to pause between HTTP POSTs
        max_tries: how many times to retry on socket.error
    Returns:
        a bool of whether the posts succeeded
    """
    if port is None:
        port = random.choice(RELAY_PORTS)
    if not isinstance(data, list):
        raise ValueError('carbon_post data must be a list')
    to_carbon = data[:]  # create a new copy in memory
    success = True
    while to_carbon and success:
        chunk = []
        for _ in range(chunk_size):
            try:
                chunk.append(to_carbon.pop(0))
            except IndexError:
                continue
        for attempt in range(max_tries):  # attempt to post it 3 times
            try:
                sock = socket.socket()
                sock.connect((server, port))
                payload = '{0}\n'.format('\n'.join(chunk))
                sock.sendall(bytes(payload, 'ascii', 'ignore'))
                sock.close()
                success = True
                break  # stop trying
            except IOError:
                success = False
                if attempt < max_tries - 1:
                    time.sleep(chunk_sleep)
        if to_carbon:
            time.sleep(chunk_sleep)
    return success



"""TypedDict typehints for Backup Authority API replies"""

from typing import TypedDict, Union, Literal
from restic import ResticRepo

# https://www.python.org/dev/peps/pep-0589/


class SharedFailoverLocks(TypedDict):
    """Return format of get_shared_failover_locks"""

    server_lock: Union[int, None]  # timestamp
    user_locks: list[str]  # snapshot IDs


class AgentClientLookup(TypedDict):
    """Reply dict format for support agent bucket lookups for v/ded servers"""

    wans: dict[str, str]
    repo: ResticRepo
    svr_class: str


class AgentCpuserLookup(TypedDict):
    """Reply dict format for support agent bucket lookups for cPanel users"""

    wans: dict[str, str]
    repo: ResticRepo


class VznodeBackupLookup(TypedDict):
    """Reply dict format for ``BakAuth.vzclient_backup``"""

    endpoints: dict[str, str]
    node_keys: ResticRepo
    vps_keys: dict[int, ResticRepo]
    # task_ids is sent by bakauth but not stored in this dict
    changed: dict[int, str]


class VznodeRestoreLookup(TypedDict):
    """Reply dict format for ``BakAuth.vzclient_restore``"""

    this_endpoint: str
    key_info: ResticRepo
    all_endpoints: dict[str, str]


class UserBuckets(TypedDict):
    """Reply dict format for ``BakAuth.get_user_buckets``"""

    copy_users: list[str]
    quotas_gb: dict[str, int]
    repos: dict[str, ResticRepo]
    missing: list[str]


class RegCopy(TypedDict):
    """Info on a cross-coast cluster"""

    name: str
    location: str
    endpoint: str


class RegDetails(TypedDict):
    """Reply dict format for ``BakAuth.get_reg_details``"""

    name: str  # cluster name
    location: str
    svr_class: Literal[
        'imh_vps',
        'imh_ded',
        'imh_reseller',
        'imh_ngx',
        'imh_shared',
        'hub_shared',
    ]
    client_host: str  # short hostname bakauth knows this client as
    endpoint: str  # url such as https://backups.lax1.imhcloud.net
    wans: dict[str, str]  # name -> urls for all clusters' wans
    copy: RegCopy
    repo: ResticRepo



'''
TaskStats module
================

All that you should know about TaskStats, is that you should not
use it. But if you have to, ok::

    import os
    from pyroute2 import TaskStats
    ts = TaskStats()
    ts.get_pid_stat(os.getpid())

It is not implemented normally yet, but some methods are already
usable.
'''

from pyroute2.netlink import NLM_F_REQUEST, genlmsg, nla, nla_struct
from pyroute2.netlink.generic import GenericNetlinkSocket

TASKSTATS_CMD_UNSPEC = 0  # Reserved
TASKSTATS_CMD_GET = 1  # user->kernel request/get-response
TASKSTATS_CMD_NEW = 2


class tcmd(genlmsg):
    nla_map = (
        ('TASKSTATS_CMD_ATTR_UNSPEC', 'none'),
        ('TASKSTATS_CMD_ATTR_PID', 'uint32'),
        ('TASKSTATS_CMD_ATTR_TGID', 'uint32'),
        ('TASKSTATS_CMD_ATTR_REGISTER_CPUMASK', 'asciiz'),
        ('TASKSTATS_CMD_ATTR_DEREGISTER_CPUMASK', 'asciiz'),
    )


class tstats(nla_struct):
    fields = (
        ('version', 'H'),  # 2
        ('ac_exitcode', 'I'),  # 4
        ('ac_flag', 'B'),  # 1
        ('ac_nice', 'B'),  # 1 --- 10
        ('cpu_count', 'Q'),  # 8
        ('cpu_delay_total', 'Q'),  # 8
        ('blkio_count', 'Q'),  # 8
        ('blkio_delay_total', 'Q'),  # 8
        ('swapin_count', 'Q'),  # 8
        ('swapin_delay_total', 'Q'),  # 8
        ('cpu_run_real_total', 'Q'),  # 8
        ('cpu_run_virtual_total', 'Q'),  # 8
        ('ac_comm', '32s'),  # 32 +++ 112
        ('ac_sched', 'B'),  # 1
        ('__ac_pad', '3x'),  # 3
        # (the ac_uid field is aligned(8), so we add more padding)
        ('__implicit_pad', '4x'),  # 4
        ('ac_uid', 'I'),  # 4  +++ 120
        ('ac_gid', 'I'),  # 4
        ('ac_pid', 'I'),  # 4
        ('ac_ppid', 'I'),  # 4
        ('ac_btime', 'I'),  # 4  +++ 136
        ('ac_etime', 'Q'),  # 8  +++ 144
        ('ac_utime', 'Q'),  # 8
        ('ac_stime', 'Q'),  # 8
        ('ac_minflt', 'Q'),  # 8
        ('ac_majflt', 'Q'),  # 8
        ('coremem', 'Q'),  # 8
        ('virtmem', 'Q'),  # 8
        ('hiwater_rss', 'Q'),  # 8
        ('hiwater_vm', 'Q'),  # 8
        ('read_char', 'Q'),  # 8
        ('write_char', 'Q'),  # 8
        ('read_syscalls', 'Q'),  # 8
        ('write_syscalls', 'Q'),  # 8
        ('read_bytes', 'Q'),  # ...
        ('write_bytes', 'Q'),
        ('cancelled_write_bytes', 'Q'),
        ('nvcsw', 'Q'),
        ('nivcsw', 'Q'),
        ('ac_utimescaled', 'Q'),
        ('ac_stimescaled', 'Q'),
        ('cpu_scaled_run_real_total', 'Q'),
    )

    def decode(self):
        nla_struct.decode(self)
        command = self['ac_comm']
        if isinstance(command, bytes):
            command = command.decode('utf-8')
        self['ac_comm'] = command[: command.find('\0')]


class taskstatsmsg(genlmsg):
    nla_map = (
        ('TASKSTATS_TYPE_UNSPEC', 'none'),
        ('TASKSTATS_TYPE_PID', 'uint32'),
        ('TASKSTATS_TYPE_TGID', 'uint32'),
        ('TASKSTATS_TYPE_STATS', 'stats'),
        ('TASKSTATS_TYPE_AGGR_PID', 'aggr_pid'),
        ('TASKSTATS_TYPE_AGGR_TGID', 'aggr_tgid'),
    )

    class stats(tstats):
        pass  # FIXME: optimize me!

    class aggr_id(nla):
        nla_map = (
            ('TASKSTATS_TYPE_UNSPEC', 'none'),
            ('TASKSTATS_TYPE_PID', 'uint32'),
            ('TASKSTATS_TYPE_TGID', 'uint32'),
            ('TASKSTATS_TYPE_STATS', 'stats'),
        )

        class stats(tstats):
            pass

    class aggr_pid(aggr_id):
        pass

    class aggr_tgid(aggr_id):
        pass


class TaskStats(GenericNetlinkSocket):
    def bind(self):
        GenericNetlinkSocket.bind(self, 'TASKSTATS', taskstatsmsg)

    def get_pid_stat(self, pid):
        '''
        Get taskstats for a process. Pid should be an integer.
        '''
        msg = tcmd()
        msg['cmd'] = TASKSTATS_CMD_GET
        msg['version'] = 1
        msg['attrs'].append(['TASKSTATS_CMD_ATTR_PID', pid])
        return self.nlm_request(msg, self.prid, msg_flags=NLM_F_REQUEST)

    def _register_mask(self, cmd, mask):
        msg = tcmd()
        msg['cmd'] = TASKSTATS_CMD_GET
        msg['version'] = 1
        msg['attrs'].append([cmd, mask])
        # there is no response to this request
        self.put(msg, self.prid, msg_flags=NLM_F_REQUEST)

    def register_mask(self, mask):
        '''
        Start the accounting for a processors by a mask. Mask is
        a string, e.g.::
            0,1 -- first two CPUs
            0-4,6-10 -- CPUs from 0 to 4 and from 6 to 10

        Though the kernel has a procedure, that cleans up accounting,
        when it is not used, it is recommended to run deregister_mask()
        before process exit.
        '''
        self._register_mask('TASKSTATS_CMD_ATTR_REGISTER_CPUMASK', mask)

    def deregister_mask(self, mask):
        '''
        Stop the accounting.
        '''
        self._register_mask('TASKSTATS_CMD_ATTR_DEREGISTER_CPUMASK', mask)



"""
Network SNMP
============

Manage the SNMP configuration on network devices.

:codeauthor: Mircea Ulinic <ping@mirceaulinic.net>
:maturity:   new
:depends:    napalm
:platform:   unix

Dependencies
------------

- :mod:`napalm snmp management module (salt.modules.napalm_snmp) <salt.modules.napalm_snmp>`

.. versionadded:: 2016.11.0
"""

import logging

import salt.utils.json
import salt.utils.napalm

log = logging.getLogger(__name__)


# ----------------------------------------------------------------------------------------------------------------------
# state properties
# ----------------------------------------------------------------------------------------------------------------------

__virtualname__ = "netsnmp"

_COMMUNITY_MODE_MAP = {
    "read-only": "ro",
    "readonly": "ro",
    "read-write": "rw",
    "write": "rw",
}

# ----------------------------------------------------------------------------------------------------------------------
# global variables
# ----------------------------------------------------------------------------------------------------------------------

# ----------------------------------------------------------------------------------------------------------------------
# property functions
# ----------------------------------------------------------------------------------------------------------------------


def __virtual__():
    """
    NAPALM library must be installed for this module to work and run in a (proxy) minion.
    """
    return salt.utils.napalm.virtual(__opts__, __virtualname__, __file__)


# ----------------------------------------------------------------------------------------------------------------------
# helper functions -- will not be exported
# ----------------------------------------------------------------------------------------------------------------------


def _ordered_dict_to_dict(config):
    """
    Forced the datatype to dict, in case OrderedDict is used.
    """

    return salt.utils.json.loads(salt.utils.json.dumps(config))


def _expand_config(config, defaults):
    """
    Completed the values of the expected config for the edge cases with the default values.
    """

    defaults.update(config)
    return defaults


def _valid_dict(dic):
    """
    Valid dictionary?
    """

    return isinstance(dic, dict) and len(dic) > 0


def _valid_str(value):
    """
    Valid str?
    """

    return isinstance(value, str) and len(value) > 0


def _community_defaults():
    """
    Returns the default values of a community.
    """

    return {"mode": "ro"}


def _clear_community_details(community_details):
    """
    Clears community details.
    """

    for key in ["acl", "mode"]:
        _str_elem(community_details, key)

    _mode = community_details.get["mode"] = community_details.get("mode").lower()

    if _mode in _COMMUNITY_MODE_MAP:
        community_details["mode"] = _COMMUNITY_MODE_MAP.get(_mode)

    if community_details["mode"] not in ["ro", "rw"]:
        community_details["mode"] = "ro"  # default is read-only

    return community_details


def _str_elem(config, key):
    """
    Re-adds the value of a specific key in the dict, only in case of valid str value.
    """

    _value = config.pop(key, "")
    if _valid_str(_value):
        config[key] = _value


def _check_config(config):
    """
    Checks the desired config and clears interesting details.
    """

    if not _valid_dict(config):
        return True, ""

    _community = config.get("community")
    _community_tmp = {}
    if not _community:
        return False, "Must specify at least a community."
    if _valid_str(_community):
        _community_tmp[_community] = _community_defaults()
    elif isinstance(_community, list):
        # if the user specifies the communities as list
        for _comm in _community:
            if _valid_str(_comm):
                # list of values
                _community_tmp[_comm] = _community_defaults()
                # default mode is read-only
            if _valid_dict(_comm):
                # list of dicts
                for _comm_name, _comm_details in _comm.items():
                    if _valid_str(_comm_name):
                        _community_tmp[_comm_name] = _clear_community_details(
                            _comm_details
                        )
    elif _valid_dict(_community):
        # directly as dict of communities
        # recommended way...
        for _comm_name, _comm_details in _community.items():
            if _valid_str(_comm_name):
                _community_tmp[_comm_name] = _clear_community_details(_comm_details)
    else:
        return False, "Please specify a community or a list of communities."

    if not _valid_dict(_community_tmp):
        return False, "Please specify at least a valid community!"

    config["community"] = _community_tmp

    for key in ["location", "contact", "chassis_id"]:
        # not mandatory, but should be here only if valid
        _str_elem(config, key)

    return True, ""


def _retrieve_device_config():
    """
    Retrieves the SNMP config from the device.
    """

    return __salt__["snmp.config"]()


def _create_diff_action(diff, diff_key, key, value):
    """
    DRY to build diff parts (added, removed, updated).
    """

    if diff_key not in diff:
        diff[diff_key] = {}
    diff[diff_key][key] = value


def _create_diff(diff, fun, key, prev, curr):
    """
    Builds the diff dictionary.
    """

    if not fun(prev):
        _create_diff_action(diff, "added", key, curr)
    elif fun(prev) and not fun(curr):
        _create_diff_action(diff, "removed", key, prev)
    elif not fun(curr):
        _create_diff_action(diff, "updated", key, curr)


def _compute_diff(existing, expected):
    """
    Computes the differences between the existing and the expected SNMP config.
    """

    diff = {}

    for key in ["location", "contact", "chassis_id"]:
        if existing.get(key) != expected.get(key):
            _create_diff(diff, _valid_str, key, existing.get(key), expected.get(key))

    for key in ["community"]:  # for the moment only onen
        if existing.get(key) != expected.get(key):
            _create_diff(diff, _valid_dict, key, existing.get(key), expected.get(key))

    return diff


def _configure(changes):
    """
    Calls the configuration template to apply the configuration changes on the device.
    """

    cfgred = True
    reasons = []
    fun = "update_config"

    for key in ["added", "updated", "removed"]:
        _updated_changes = changes.get(key, {})
        if not _updated_changes:
            continue
        _location = _updated_changes.get("location", "")
        _contact = _updated_changes.get("contact", "")
        _community = _updated_changes.get("community", {})
        _chassis_id = _updated_changes.get("chassis_id", "")
        if key == "removed":
            fun = "remove_config"
        _ret = __salt__[f"snmp.{fun}"](
            location=_location,
            contact=_contact,
            community=_community,
            chassis_id=_chassis_id,
            commit=False,
        )
        cfgred = cfgred and _ret.get("result")
        if not _ret.get("result") and _ret.get("comment"):
            reasons.append(_ret.get("comment"))

    return {"result": cfgred, "comment": "\n".join(reasons) if reasons else ""}


# ----------------------------------------------------------------------------------------------------------------------
# callable functions
# ----------------------------------------------------------------------------------------------------------------------


def managed(name, config=None, defaults=None):
    """
    Configures the SNMP on the device as specified in the SLS file.

    SLS Example:

    .. code-block:: yaml

        snmp_example:
            netsnmp.managed:
                 - config:
                    location: Honolulu, HI, US
                 - defaults:
                    contact: noc@cloudflare.com

    Output example (for the SLS above, e.g. called snmp.sls under /router/):

    .. code-block:: bash

        $ sudo salt edge01.hnl01 state.sls router.snmp test=True
        edge01.hnl01:
        ----------
                  ID: snmp_example
            Function: snmp.managed
              Result: None
             Comment: Testing mode: configuration was not changed!
             Started: 13:29:06.872363
            Duration: 920.466 ms
             Changes:
                      ----------
                      added:
                          ----------
                          chassis_id:
                              None
                          contact:
                              noc@cloudflare.com
                          location:
                              Honolulu, HI, US

        Summary for edge01.hnl01
        ------------
        Succeeded: 1 (unchanged=1, changed=1)
        Failed:    0
        ------------
        Total states run:     1
        Total run time: 920.466 ms
    """

    result = False
    comment = ""
    changes = {}

    ret = {"name": name, "changes": changes, "result": result, "comment": comment}

    # make sure we're working only with dict
    config = _ordered_dict_to_dict(config)
    defaults = _ordered_dict_to_dict(defaults)

    expected_config = _expand_config(config, defaults)
    if not isinstance(expected_config, dict):
        ret["comment"] = "User provided an empty SNMP config!"
        return ret
    valid, message = _check_config(expected_config)

    if not valid:  # check and clean
        ret["comment"] = "Please provide a valid configuration: {error}".format(
            error=message
        )
        return ret

    # ----- Retrieve existing users configuration and determine differences ------------------------------------------->

    _device_config = _retrieve_device_config()
    if not _device_config.get("result"):
        ret["comment"] = "Cannot retrieve SNMP config from the device: {reason}".format(
            reason=_device_config.get("comment")
        )
        return ret

    device_config = _device_config.get("out", {})

    if device_config == expected_config:
        ret.update({"comment": "SNMP already configured as needed.", "result": True})
        return ret

    diff = _compute_diff(device_config, expected_config)

    changes.update(diff)

    ret.update({"changes": changes})

    if __opts__["test"] is True:
        ret.update(
            {"result": None, "comment": "Testing mode: configuration was not changed!"}
        )
        return ret

    # <---- Retrieve existing NTP peers and determine peers to be added/removed --------------------------------------->

    # ----- Call _set_users and _delete_users as needed ------------------------------------------------------->

    expected_config_change = False
    result = True

    if diff:
        _configured = _configure(diff)
        if _configured.get("result"):
            expected_config_change = True
        else:  # something went wrong...
            result = False
            comment = (
                "Cannot push new SNMP config: \n{reason}".format(
                    reason=_configured.get("comment")
                )
                + comment
            )

    if expected_config_change:
        result, comment = __salt__["net.config_control"]()

    # <---- Call _set_users and _delete_users as needed --------------------------------------------------------

    ret.update({"result": result, "comment": comment})

    return ret



"""
Read tops data from a mongodb collection

This module will load tops data from a mongo collection. It uses the node's id
for lookups.

Salt Master Mongo Configuration
===============================

The module shares the same base mongo connection variables as
:py:mod:`salt.returners.mongo_return`. These variables go in your master
config file.

   * ``mongo.db`` - The mongo database to connect to. Defaults to ``'salt'``.
   * ``mongo.host`` - The mongo host to connect to. Supports replica sets by
     specifying all hosts in the set, comma-delimited. Defaults to ``'salt'``.
   * ``mongo.port`` - The port that the mongo database is running on. Defaults
     to ``27017``.
   * ``mongo.user`` - The username for connecting to mongo. Only required if
     you are using mongo authentication. Defaults to ``''``.
   * ``mongo.password`` - The password for connecting to mongo. Only required
     if you are using mongo authentication. Defaults to ``''``.


Configuring the Mongo Tops Subsystem
====================================

.. code-block:: yaml

  master_tops:
    mongo:
      collection: tops
      id_field: _id
      re_replace: ""
      re_pattern: \\.example\\.com
      states_field: states
      environment_field: environment


Module Documentation
====================
"""

import logging
import re

try:
    import pymongo

    HAS_PYMONGO = True
except ImportError:
    HAS_PYMONGO = False


__opts__ = {
    "mongo.db": "salt",
    "mongo.host": "salt",
    "mongo.password": "",
    "mongo.port": 27017,
    "mongo.user": "",
}


def __virtual__():
    if not HAS_PYMONGO:
        return False
    return "mongo"


# Set up logging
log = logging.getLogger(__name__)


def top(**kwargs):
    """
    Connect to a mongo database and read per-node tops data.

    Parameters:
        * `collection`: The mongodb collection to read data from. Defaults to
          ``'tops'``.
        * `id_field`: The field in the collection that represents an individual
          minion id. Defaults to ``'_id'``.
        * `re_pattern`: If your naming convention in the collection is shorter
          than the minion id, you can use this to trim the name.
          `re_pattern` will be used to match the name, and `re_replace` will
          be used to replace it. Backrefs are supported as they are in the
          Python standard library. If ``None``, no mangling of the name will
          be performed - the collection will be searched with the entire
          minion id. Defaults to ``None``.
        * `re_replace`: Use as the replacement value in node ids matched with
          `re_pattern`. Defaults to ''. Feel free to use backreferences here.
        * `states_field`: The name of the field providing a list of states.
        * `environment_field`: The name of the field providing the environment.
          Defaults to ``environment``.
    """
    host = __opts__["mongo.host"]
    port = __opts__["mongo.port"]
    collection = __opts__["master_tops"]["mongo"].get("collection", "tops")
    id_field = __opts__["master_tops"]["mongo"].get("id_field", "_id")
    re_pattern = __opts__["master_tops"]["mongo"].get("re_pattern", "")
    re_replace = __opts__["master_tops"]["mongo"].get("re_replace", "")
    states_field = __opts__["master_tops"]["mongo"].get("states_field", "states")
    environment_field = __opts__["master_tops"]["mongo"].get(
        "environment_field", "environment"
    )

    log.info("connecting to %s:%s for mongo ext_tops", host, port)
    conn = pymongo.MongoClient(host, port)

    log.debug("using database '%s'", __opts__["mongo.db"])
    mdb = conn[__opts__["mongo.db"]]

    user = __opts__.get("mongo.user")
    password = __opts__.get("mongo.password")

    if user and password:
        log.debug("authenticating as '%s'", user)
        mdb.authenticate(user, password)

    # Do the regex string replacement on the minion id
    minion_id = kwargs["opts"]["id"]
    if re_pattern:
        minion_id = re.sub(re_pattern, re_replace, minion_id)

    log.info(
        "ext_tops.mongo: looking up tops def for {'%s': '%s'} in mongo",
        id_field,
        minion_id,
    )

    result = mdb[collection].find_one(
        {id_field: minion_id}, projection=[states_field, environment_field]
    )
    if result and states_field in result:
        if environment_field in result:
            environment = result[environment_field]
        else:
            environment = "base"
        log.debug("ext_tops.mongo: found document, returning states")
        return {environment: result[states_field]}
    else:
        # If we can't find the minion the database it's not necessarily an
        # error.
        log.debug("ext_tops.mongo: no document found in collection %s", collection)
        return {}



"""
Genshi Renderer for Salt
"""

try:
    from genshi.template import MarkupTemplate  # pylint: disable=no-name-in-module
    from genshi.template import NewTextTemplate  # pylint: disable=no-name-in-module
    from genshi.template import OldTextTemplate  # pylint: disable=no-name-in-module

    HAS_LIBS = True
except ImportError:
    HAS_LIBS = False


def render(genshi_data, saltenv="base", sls="", method="xml", **kws):
    """
    Render a Genshi template. A method should be passed in as part of the
    kwargs. If no method is passed in, xml is assumed. Valid methods are:

    .. code-block:

        - xml
        - xhtml
        - html
        - text
        - newtext
        - oldtext

    Note that the ``text`` method will call ``NewTextTemplate``. If ``oldtext``
    is desired, it must be called explicitly

    :rtype: A Python data structure
    """
    if not HAS_LIBS:
        return {}

    if not isinstance(genshi_data, str):
        genshi_data = genshi_data.read()

    if genshi_data.startswith("#!"):
        genshi_data = genshi_data[(genshi_data.find("\n") + 1) :]
    if not genshi_data.strip():
        return {}

    if method == "text" or method == "newtext":
        tmpl = NewTextTemplate(genshi_data)
    elif method == "oldtext":
        tmpl = OldTextTemplate(genshi_data)
    else:
        tmpl = MarkupTemplate(genshi_data)

    return tmpl.generate(**kws).render(method)



"""
Netmiko Execution Module
========================

.. versionadded:: 2019.2.0

Execution module to interface the connection with a remote network device. It is
flexible enough to execute the commands both when running under a Netmiko Proxy
Minion, as well as running under a Regular Minion by specifying the connection
arguments, i.e., ``device_type``, ``ip``, ``username``, ``password`` etc.

:codeauthor: Mircea Ulinic <ping@mirceaulinic.net> & Kirk Byers <ktbyers@twb-tech.com>
:maturity:   new
:depends:    netmiko
:platform:   unix

Dependencies
------------

The ``netmiko`` proxy modules requires Netmiko to be installed: ``pip install netmiko``.

Usage
-----

This module can equally be used via the :mod:`netmiko <salt.proxy.netmiko_px>`
Proxy module (check documentation), or directly from an arbitrary (Proxy) Minion
that is running on a server (computer) having access to the network device, and
has the ``netmiko`` library installed.

When running outside of the :mod:`netmiko Proxy <salt.proxy.netmiko_px>` (i.e.,
from another Proxy Minion type, or regular Minion), the netmiko connection
arguments can be either specified from the CLI when executing the command, or
in a configuration block under the ``netmiko`` key in the configuration opts
(i.e., (Proxy) Minion configuration file), or Pillar. The module supports these
simultaneously. These fields are the exact same supported by the ``netmiko``
Proxy Module:

- ``device_type`` - Class selection based on device type. Supported options:

  - ``a10``: A10 Networks
  - ``accedian``: Accedian Networks
  - ``alcatel_aos``: Alcatel AOS
  - ``alcatel_sros``: Alcatel SROS
  - ``apresia_aeos``: Apresia AEOS
  - ``arista_eos``: Arista EOS
  - ``aruba_os``: Aruba
  - ``avaya_ers``: Avaya ERS
  - ``avaya_vsp``: Avaya VSP
  - ``brocade_fastiron``: Brocade Fastiron
  - ``brocade_netiron``: Brocade Netiron
  - ``brocade_nos``: Brocade NOS
  - ``brocade_vdx``: Brocade NOS
  - ``brocade_vyos``: VyOS
  - ``checkpoint_gaia``: Check Point GAiA
  - ``calix_b6``: Calix B6
  - ``ciena_saos``: Ciena SAOS
  - ``cisco_asa``: Cisco SA
  - ``cisco_ios``: Cisco IOS
  - ``cisco_nxos``: Cisco NX-oS
  - ``cisco_s300``: Cisco S300
  - ``cisco_tp``: Cisco TpTcCe
  - ``cisco_wlc``: Cisco WLC
  - ``cisco_xe``: Cisco IOS
  - ``cisco_xr``: Cisco XR
  - ``coriant``: Coriant
  - ``dell_force10``: Dell Force10
  - ``dell_os10``: Dell OS10
  - ``dell_powerconnect``: Dell PowerConnect
  - ``eltex``: Eltex
  - ``enterasys``: Enterasys
  - ``extreme``: Extreme
  - ``extreme_wing``: Extreme Wing
  - ``f5_ltm``: F5 LTM
  - ``fortinet``: Fortinet
  - ``generic_termserver``: TerminalServer
  - ``hp_comware``: HP Comware
  - ``hp_procurve``: HP Procurve
  - ``huawei``: Huawei
  - ``huawei_vrpv8``: Huawei VRPV8
  - ``juniper``: Juniper Junos
  - ``juniper_junos``: Juniper Junos
  - ``linux``: Linux
  - ``mellanox``: Mellanox
  - ``mrv_optiswitch``: MrvOptiswitch
  - ``netapp_cdot``: NetAppcDot
  - ``netscaler``: Netscaler
  - ``ovs_linux``: OvsLinux
  - ``paloalto_panos``: PaloAlto Panos
  - ``pluribus``: Pluribus
  - ``quanta_mesh``: Quanta Mesh
  - ``ruckus_fastiron``: Ruckus Fastiron
  - ``ubiquiti_edge``: Ubiquiti Edge
  - ``ubiquiti_edgeswitch``: Ubiquiti Edge
  - ``vyatta_vyos``: VyOS
  - ``vyos``: VyOS
  - ``brocade_fastiron_telnet``: Brocade Fastiron over Telnet
  - ``brocade_netiron_telnet``: Brocade Netiron over Telnet
  - ``cisco_ios_telnet``: Cisco IOS over Telnet
  - ``apresia_aeos_telnet``: Apresia AEOS over Telnet
  - ``arista_eos_telnet``: Arista EOS over Telnet
  - ``hp_procurve_telnet``: HP Procurve over Telnet
  - ``hp_comware_telnet``: HP Comware over Telnet
  - ``juniper_junos_telnet``: Juniper Junos over Telnet
  - ``calix_b6_telnet``: Calix B6 over Telnet
  - ``dell_powerconnect_telnet``: Dell PowerConnect over Telnet
  - ``generic_termserver_telnet``: TerminalServer over Telnet
  - ``extreme_telnet``: Extreme Networks over Telnet
  - ``ruckus_fastiron_telnet``: Ruckus Fastiron over Telnet
  - ``cisco_ios_serial``: Cisco IOS over serial port

- ``ip`` - IP address of target device (not required if ``host`` is provided)

- ``host`` - Hostname of target device (not required if ``ip`` is provided)

- ``username`` - Username to authenticate against target device, if required

- ``password`` - Password to authenticate against target device, if required

- ``secret`` - The enable password if target device requires one

- ``port`` - The destination port used to connect to the target device

- ``global_delay_factor`` - Multiplication factor affecting Netmiko delays
  (default: ``1``)

- ``use_keys`` - Connect to target device using SSH keys (default: ``False``)

- ``key_file`` - Filename path of the SSH key file to use

- ``allow_agent`` - Enable use of SSH key-agent

- ``ssh_strict`` - Automatically reject unknown SSH host keys (default:
  ``False``, which means unknown SSH host keys will be accepted)

- ``system_host_keys`` - Load host keys from the user's "known_hosts" file
  (default: ``False``)

- ``alt_host_keys`` - If ``True``,  host keys will be loaded from the file
  specified in ``alt_key_file`` (default: ``False``)

- ``alt_key_file`` - SSH host key file to use (if ``alt_host_keys=True``)

- ``ssh_config_file`` - File name of OpenSSH configuration file

- ``timeout`` - Connection timeout, in seconds (default: ``90``)

- ``session_timeout`` - Set a timeout for parallel requests, in seconds
  (default: ``60``)

- ``keepalive`` - Send SSH keepalive packets at a specific interval, in
  seconds. Currently defaults to ``0``, for backwards compatibility (it will
  not attempt to keep the connection alive using the KEEPALIVE packets).

- ``default_enter`` - Character(s) to send to correspond to enter key (default:
  ``\\n``)

- ``response_return`` - Character(s) to use in normalized return data to
  represent enter key (default: ``\\n``)

Example (when not running in a ``netmiko`` Proxy Minion):

.. code-block:: yaml

  netmiko:
    username: test
    password: test

In case the ``username`` and ``password`` are the same on any device you are
targeting, the block above (besides other parameters specific to your
environment you might need) should suffice to be able to execute commands from
outside a ``netmiko`` Proxy, e.g.:

.. code-block:: bash

    salt '*' netmiko.send_command 'show version' host=router1.example.com device_type=juniper
    salt '*' netmiko.send_config https://bit.ly/2sgljCB host=sw2.example.com device_type=cisco_ios

.. note::

    Remember that the above applies only when not running in a ``netmiko`` Proxy
    Minion. If you want to use the :mod:`<salt.proxy.netmiko_px>`, please follow
    the documentation notes for a proper setup.
"""

import logging

import salt.utils.platform
from salt.exceptions import CommandExecutionError
from salt.utils.args import clean_kwargs

try:
    from netmiko import BaseConnection, ConnectHandler

    HAS_NETMIKO = True
except ImportError:
    HAS_NETMIKO = False

# -----------------------------------------------------------------------------
# execution module properties
# -----------------------------------------------------------------------------

__proxyenabled__ = ["*"]
# Any Proxy Minion should be able to execute these (not only netmiko)

__virtualname__ = "netmiko"
# The Execution Module will be identified as ``netmiko``

# -----------------------------------------------------------------------------
# globals
# -----------------------------------------------------------------------------

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------
# propery functions
# -----------------------------------------------------------------------------


def __virtual__():
    """
    Execution module available only if Netmiko is installed.
    """
    if not HAS_NETMIKO:
        return (
            False,
            "The netmiko execution module requires netmiko library to be installed.",
        )
    if (
        salt.utils.platform.is_proxy()
        and __opts__["proxy"]["proxytype"] == "deltaproxy"
    ):
        return (
            False,
            "Unsupported proxy minion type.",
        )

    return __virtualname__


# -----------------------------------------------------------------------------
# helper functions
# -----------------------------------------------------------------------------


def _prepare_connection(**kwargs):
    """
    Prepare the connection with the remote network device, and clean up the key
    value pairs, removing the args used for the connection init.
    """
    init_args = {}
    fun_kwargs = {}
    netmiko_kwargs = __salt__["config.get"]("netmiko", {})
    netmiko_kwargs.update(kwargs)  # merge the CLI args with the opts/pillar
    netmiko_init_args, _, _, netmiko_defaults = __utils__["args.get_function_argspec"](
        BaseConnection.__init__
    )
    check_self = netmiko_init_args.pop(0)
    for karg, warg in netmiko_kwargs.items():
        if karg not in netmiko_init_args:
            if warg is not None:
                fun_kwargs[karg] = warg
            continue
        if warg is not None:
            init_args[karg] = warg
    conn = ConnectHandler(**init_args)
    return conn, fun_kwargs


# -----------------------------------------------------------------------------
# callable functions
# -----------------------------------------------------------------------------


def get_connection(**kwargs):
    """
    Return the Netmiko connection object.

    .. warning::

        This function returns an unserializable object, hence it is not meant
        to be used on the CLI. This should mainly be used when invoked from
        other modules for the low level connection with the network device.

    kwargs
        Key-value dictionary with the authentication details.

    USAGE Example:

    .. code-block:: python

        conn = __salt__['netmiko.get_connection'](host='router1.example.com',
                                                  username='example',
                                                  password='example')
        show_if = conn.send_command('show interfaces')
        conn.disconnect()
    """
    kwargs = clean_kwargs(**kwargs)
    if "netmiko.conn" in __proxy__:
        return __proxy__["netmiko.conn"]()
    conn, kwargs = _prepare_connection(**kwargs)
    return conn


def call(method, *args, **kwargs):
    """
    Invoke an arbitrary Netmiko method.

    method
        The name of the Netmiko method to invoke.

    args
        A list of arguments to send to the method invoked.

    kwargs
        Key-value dictionary to send to the method invoked.
    """
    kwargs = clean_kwargs(**kwargs)
    if "netmiko.call" in __proxy__:
        return __proxy__["netmiko.call"](method, *args, **kwargs)
    conn, kwargs = _prepare_connection(**kwargs)
    ret = getattr(conn, method)(*args, **kwargs)
    conn.disconnect()
    return ret


def multi_call(*methods, **kwargs):
    """
    Invoke multiple Netmiko methods at once, and return their output, as list.

    methods
        A list of dictionaries with the following keys:

        - ``name``: the name of the Netmiko method to be executed.
        - ``args``: list of arguments to be sent to the Netmiko method.
        - ``kwargs``: dictionary of arguments to be sent to the Netmiko method.

    kwargs
        Key-value dictionary with the connection details (when not running
        under a Proxy Minion).
    """
    kwargs = clean_kwargs(**kwargs)
    if "netmiko.conn" in __proxy__:
        conn = __proxy__["netmiko.conn"]()
    else:
        conn, kwargs = _prepare_connection(**kwargs)
    ret = []
    for method in methods:
        # Explicit unpacking
        method_name = method["name"]
        method_args = method.get("args", [])
        method_kwargs = method.get("kwargs", [])
        ret.append(getattr(conn, method_name)(*method_args, **method_kwargs))
    if "netmiko.conn" not in __proxy__:
        conn.disconnect()
    return ret


def send_command(command_string, **kwargs):
    """
    Execute command_string on the SSH channel using a pattern-based mechanism.
    Generally used for show commands. By default this method will keep waiting
    to receive data until the network device prompt is detected. The current
    network device prompt will be determined automatically.

    command_string
        The command to be executed on the remote device.

    expect_string
        Regular expression pattern to use for determining end of output.
        If left blank will default to being based on router prompt.

    delay_factor: ``1``
        Multiplying factor used to adjust delays (default: ``1``).

    max_loops: ``500``
        Controls wait time in conjunction with delay_factor. Will default to be
        based upon self.timeout.

    auto_find_prompt: ``True``
        Whether it should try to auto-detect the prompt (default: ``True``).

    strip_prompt: ``True``
        Remove the trailing router prompt from the output (default: ``True``).

    strip_command: ``True``
        Remove the echo of the command from the output (default: ``True``).

    normalize: ``True``
        Ensure the proper enter is sent at end of command (default: ``True``).

    use_textfsm: ``False``
        Process command output through TextFSM template (default: ``False``).

    CLI Example:

    .. code-block:: bash

        salt '*' netmiko.send_command 'show version'
        salt '*' netmiko.send_command 'show_version' host='router1.example.com' username='example' device_type='cisco_ios'
    """
    return call("send_command", command_string, **kwargs)


def send_command_timing(command_string, **kwargs):
    """
    Execute command_string on the SSH channel using a delay-based mechanism.
    Generally used for show commands.

    command_string
        The command to be executed on the remote device.

    delay_factor: ``1``
        Multiplying factor used to adjust delays (default: ``1``).

    max_loops: ``500``
        Controls wait time in conjunction with delay_factor. Will default to be
        based upon self.timeout.

    strip_prompt: ``True``
        Remove the trailing router prompt from the output (default: ``True``).

    strip_command: ``True``
        Remove the echo of the command from the output (default: ``True``).

    normalize: ``True``
        Ensure the proper enter is sent at end of command (default: ``True``).

    use_textfsm: ``False``
        Process command output through TextFSM template (default: ``False``).

    CLI Example:

    .. code-block:: bash

        salt '*' netmiko.send_command_timing 'show version'
        salt '*' netmiko.send_command_timing 'show version' host='router1.example.com' username='example' device_type='arista_eos'
    """
    return call("send_command_timing", command_string, **kwargs)


def enter_config_mode(**kwargs):
    """
    Enter into config mode.

    config_command
        Configuration command to send to the device.

    pattern
        Pattern to terminate reading of channel.

    CLI Example:

    .. code-block:: bash

        salt '*' netmiko.enter_config_mode
        salt '*' netmiko.enter_config_mode device_type='juniper_junos' ip='192.168.0.1' username='example'
    """
    return call("config_mode", **kwargs)


def exit_config_mode(**kwargs):
    """
    Exit from configuration mode.

    exit_config
        Command to exit configuration mode.

    pattern
        Pattern to terminate reading of channel.

    CLI Example:

    .. code-block:: bash

        salt '*' netmiko.exit_config_mode
        salt '*' netmiko.exit_config_mode device_type='juniper' ip='192.168.0.1' username='example'
    """
    return call("exit_config_mode", **kwargs)


def send_config(
    config_file=None,
    config_commands=None,
    template_engine="jinja",
    commit=False,
    context=None,
    defaults=None,
    saltenv="base",
    **kwargs,
):
    """
    Send configuration commands down the SSH channel.
    Return the configuration lines sent to the device.

    The function is flexible to send the configuration from a local or remote
    file, or simply the commands as list.

    config_file
        The source file with the configuration commands to be sent to the
        device.

        The file can also be a template that can be rendered using the template
        engine of choice.

        This can be specified using the absolute path to the file, or using one
        of the following URL schemes:

        - ``salt://``, to fetch the file from the Salt fileserver.
        - ``http://`` or ``https://``
        - ``ftp://``
        - ``s3://``
        - ``swift://``

    config_commands
        Multiple configuration commands to be sent to the device.

        .. note::

            This argument is ignored when ``config_file`` is specified.

    template_engine: ``jinja``
        The template engine to use when rendering the source file. Default:
        ``jinja``. To simply fetch the file without attempting to render, set
        this argument to ``None``.

    commit: ``False``
        Commit the configuration changes before exiting the config mode. This
        option is by default disabled, as many platforms don't have this
        capability natively.

    context
        Variables to add to the template context.

    defaults
        Default values of the context_dict.

    exit_config_mode: ``True``
        Determines whether or not to exit config mode after complete.

    delay_factor: ``1``
        Factor to adjust delays.

    max_loops: ``150``
        Controls wait time in conjunction with delay_factor (default: ``150``).

    strip_prompt: ``False``
        Determines whether or not to strip the prompt (default: ``False``).

    strip_command: ``False``
        Determines whether or not to strip the command (default: ``False``).

    config_mode_command
        The command to enter into config mode.

    CLI Example:

    .. code-block:: bash

        salt '*' netmiko.send_config config_commands="['interface GigabitEthernet3', 'no ip address']"
        salt '*' netmiko.send_config config_commands="['snmp-server location {{ grains.location }}']"
        salt '*' netmiko.send_config config_file=salt://config.txt
        salt '*' netmiko.send_config config_file=https://bit.ly/2sgljCB device_type='cisco_ios' ip='1.2.3.4' username='example'
    """
    if config_file:
        file_str = __salt__["cp.get_file_str"](config_file, saltenv=saltenv)
        if file_str is False:
            raise CommandExecutionError(f"Source file {config_file} not found")
    elif config_commands:
        if isinstance(config_commands, ((str,), str)):
            config_commands = [config_commands]
        file_str = "\n".join(config_commands)
        # unify all the commands in a single file, to render them in a go
    if template_engine:
        file_str = __salt__["file.apply_template_on_contents"](
            file_str, template_engine, context, defaults, saltenv
        )
    # whatever the source of the commands would be, split them line by line
    config_commands = [line for line in file_str.splitlines() if line.strip()]
    kwargs = clean_kwargs(**kwargs)
    if "netmiko.conn" in __proxy__:
        conn = __proxy__["netmiko.conn"]()
        if not conn or not conn.is_alive():
            conn, _ = _prepare_connection(**__proxy__["netmiko.args"]())
    else:
        conn, kwargs = _prepare_connection(**kwargs)
    if commit:
        kwargs["exit_config_mode"] = False  # don't exit config mode after
        # loading the commands, wait for explicit commit
    ret = conn.send_config_set(config_commands=config_commands, **kwargs)
    if commit:
        ret += conn.commit()
    return ret


def commit(**kwargs):
    """
    Commit the configuration changes.

    .. warning::

        This function is supported only on the platforms that support the
        ``commit`` operation.

    CLI Example:

    .. code-block:: bash

        salt '*' netmiko.commit
    """
    return call("commit", **kwargs)
